#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
import os
from six import text_type as _text_type
from mmdnn.conversion.examples.imagenet_test import TestKit
import torch
import torchvision.models as models


NETWORKS_MAP = {
    'inception_v3'      : lambda : models.inception_v3(pretrained=True),
    'vgg16'             : lambda : models.vgg16(pretrained=True),
    'vgg19'             : lambda : models.vgg19(pretrained=True),
    'resnet152'         : lambda : models.resnet152(pretrained=True),
    'densenet'          : lambda : models.densenet201(pretrained=True),
    'squeezenet'        : lambda : models.squeezenet1_1(pretrained=True)
}


def _main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-n', '--network',
                        type=_text_type, help='Model Type', required=True,
                        choices=NETWORKS_MAP.keys())

    parser.add_argument('-i', '--image', type=_text_type, help='Test Image Path')

    args = parser.parse_args()

    file_name = "imagenet_{}.pth".format(args.network)
    if not os.path.exists(file_name):
        model = NETWORKS_MAP.get(args.network)
        model = model()
        torch.save(model, file_name)
        print("PyTorch pretrained model is saved as [{}].".format(file_name))
    else:
        print("File [{}] existed!".format(file_name))
        model = torch.load(file_name)

    if args.image:
        import numpy as np
        func = TestKit.preprocess_func['pytorch'][args.network]
        img = func(args.image)
        img = np.transpose(img, (2, 0, 1))
        img = np.expand_dims(img, 0).copy()
        data = torch.from_numpy(img)
        data = torch.autograd.Variable(data, requires_grad=False)

        model.eval()
        predict = model(data).data.numpy()
        predict = np.squeeze(predict)
        top_indices = predict.argsort()[-5:][::-1]
        result = [(i, predict[i]) for i in top_indices]
        print(result)

        # layer_name = 'block2_pool'
        # intermediate_layer_model = keras.Model(inputs=model.input,
        #                                  outputs=model.get_layer(layer_name).output)
        # intermediate_output = intermediate_layer_model.predict(img)
        # print (intermediate_output)
        # print (intermediate_output.shape)
        # print ("%.30f" % np.sum(intermediate_output))


if __name__ == '__main__':
    _main()



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
import os
from mmdnn.conversion.examples.darknet import darknet as cdarknet
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.examples.extractor import base_extractor
from mmdnn.conversion.common.utils import download_file


class darknet_extractor(base_extractor):

    _base_model_url = "https://raw.githubusercontent.com/pjreddie/darknet/master/"

    architecture_map = {
        'yolov3'          : {
            'config'           : _base_model_url + "cfg/yolov3.cfg",
            'weights'          : "https://pjreddie.com/media/files/yolov3.weights"
        },

        'yolov2'          :{
            'config'           : _base_model_url + "cfg/yolov2.cfg",
            'weights'          : "https://pjreddie.com/media/files/yolov2.weights"
        }

    }


    @classmethod
    def download(cls, architecture, path = './'):

        if cls.sanity_check(architecture):
            cfg_name = architecture + ".cfg"
            architecture_file = download_file(cls.architecture_map[architecture]['config'], directory=path, local_fname=cfg_name)
            if not architecture_file:
                return None

            weight_name = architecture + ".weights"
            weight_file = download_file(cls.architecture_map[architecture]['weights'], directory=path, local_fname=weight_name)
            if not weight_file:
                return None

            print("Darknet Model {} saved as [{}] and [{}].".format(architecture, architecture_file, weight_file))
            return (architecture_file, weight_file)

        else:
            return None


    @classmethod
    def inference(cls, architecture, files, model_path, image_path):
        import numpy as np

        if cls.sanity_check(architecture):
            download_file(cls._base_model_url + "cfg/coco.data", directory='./')
            download_file(cls._base_model_url + "data/coco.names", directory='./data/')

            print(files)
            net = cdarknet.load_net(files[0].encode(), files[1].encode(), 0)
            meta = cdarknet.load_meta("coco.data".encode())


            r = cdarknet.detect(net, meta, image_path.encode())
            # print(r)
            return r

        else:
            return None



# d = darknet_extractor()
# model_filename = d.download('yolov3')
# print(model_filename)

# image_path = "./mmdnn/conversion/examples/data/dog.jpg"
# model_path = "./"
# d = darknet_extractor()
# result = d.inference('yolov3', model_filename, model_path, image_path = image_path)
# print(result)



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import division

import os
import numpy as np
from six import string_types as _string_types
from mmdnn.conversion.common.IR.IR_graph import IRGraph, IRGraphNode
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.DataStructure.emitter import Emitter
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.coreml.coreml_utils import _infer_coreml_input_shape

from coremltools.models.neural_network import NeuralNetworkBuilder as _NeuralNetworkBuilder
from coremltools.models import datatypes
from coremltools.models import MLModel as _MLModel
from coremltools.models.utils import save_spec as _save_spec


class CoreMLEmitter(Emitter):

    def __init__(self, architecture, weight):
        super(CoreMLEmitter, self).__init__()
        if os.path.exists(architecture) == False:
            raise ValueError("IR architecture file [{}] is not found.".format(architecture))
        else:
            self.IR_graph = IRGraph(architecture)
            self.IR_graph.build()

        if os.path.exists(weight) == False:
            raise ValueError("IR weight file [{}] is not found.".format(weight))
        else:
            self._load_weights(weight)


    def _get_inout(self):
        input_features = []
        output_features = []
        for input_node in self.IR_graph.input_layers:
            if self.IR_graph.get_node(input_node).type == 'Const':
                continue
            shape = shape_to_list(self.IR_graph.get_node(input_node).get_attr('shape'))
            shape = _infer_coreml_input_shape(shape)
            input_features.append((str(input_node), shape))
            print("CoreML Model Input Layer: [{}] {}".format(input_node, shape))

        for output_node in self.IR_graph.output_layers:

            node = self.IR_graph.get_node(output_node)

            if node.type == 'Pack':
                continue

            node.out_edges.append(node.name)
            shape = node.get_attr('_output_shapes')
            if shape:
                shape = shape_to_list(shape[0])
            else:
                shape = [1]

            shape = _infer_coreml_input_shape(shape)

            output_features.append((str(output_node), shape))

            print("CoreML Model Output Layer: [{}] {}".format(output_node, shape))


        return list(input_features), list(output_features)

    def _connect_coreml_layers(self):
        for layer in self.builder.nn_spec.layers:
            for i, out_node in enumerate(layer.output):
                layer.output[i] = self.IR_graph.get_node(out_node).real_name

    def gen_model(self,
                  input_names=None,
                  output_names=None,
                  image_input_names=None,
                  is_bgr=False,
                  red_bias=0.0,
                  green_bias=0.0,
                  blue_bias=0.0,
                  gray_bias=0.0,
                  image_scale=1.0,
                  class_labels=None,
                  predicted_feature_name=None,
                  predicted_probabilities_output=''):

        input_features, output_features = self._get_inout()
        is_classifier = class_labels is not None
        mode = 'classifier' if is_classifier else None
        self.builder = _NeuralNetworkBuilder(input_features, output_features, mode=mode)

        for layer in self.IR_graph.topological_sort:
            current_node = self.IR_graph.get_node(layer)
            print("Converting layer {}({})".format(current_node.name, current_node.type))
            node_type = current_node.type
            if hasattr(self, "emit_" + node_type):
                func = getattr(self, "emit_" + node_type)
                func(current_node)
            else:
                print("CoreMLEmitter has not supported operator [%s]." % (node_type))
                self.emit_UNKNOWN(current_node)
                assert False

        # Add classifier classes (if applicable)
        if is_classifier:
            classes_in = class_labels
            if isinstance(classes_in, _string_types):
                if not os.path.isfile(classes_in):
                    raise ValueError("Path to class labels [{}] does not exist.".format(classes_in))
                with open(classes_in, 'r') as f:
                    classes = f.read()
                classes = classes.splitlines()
            elif type(classes_in) is list: # list[int or str]
                classes = classes_in
            else:
                raise ValueError('Class labels must be a list of integers / strings, or a file path')

            if predicted_feature_name is not None:
                self.builder.set_class_labels(classes, predicted_feature_name = predicted_feature_name,
                    prediction_blob = predicted_probabilities_output)
            else:
                self.builder.set_class_labels(classes)

        # Set pre-processing paramsters
        self.builder.set_pre_processing_parameters(
            image_input_names=[input_features[0][0]],
            #image_input_names,
            is_bgr=is_bgr,
            red_bias=red_bias,
            green_bias=green_bias,
            blue_bias=blue_bias,
            gray_bias=gray_bias,
            image_scale=image_scale)

        # Return the protobuf spec
        # model = _MLModel(self.builder.spec)

        print (self.builder.spec.description)

        return self.builder.spec, input_features, output_features


    @staticmethod
    def _get_padding(IR_node):

        auto_pad = IR_node.get_attr('auto_pad')
        if auto_pad is not None:
            if auto_pad == 'VALID':
                pass
            else:
                return 'SAME'

        pads = IR_node.get_attr('pads', [0,0,0,0,0,0,0,0])

        return pads

    def emit_Mul(self, IR_node):
        """
        Not implement yet
        """
        pass
        # if IR_node.name in self.weights_dict and 'weights' in self.weights_dict[IR_node.name]:
        #     pass
        
        # self._emit_merge(IR_node,'DOT')
        

    def _emit_merge(self, IR_node, func):
        """
        Convert concat layer to coreml.
        """
        # Get input and output names
        input_names = [self.IR_graph.get_node(inp).real_name for inp in IR_node.in_edges]

        self.builder.add_elementwise(name=IR_node.name, input_names=input_names,
            output_name=IR_node.name, mode=func)

    def emit_Conv(self, IR_node):
        """
        Convert convolution layer to coreml.
        """
        has_bias = IR_node.get_attr('use_bias', False)
        is_deconv = False


        # Dimensions and weights
        kernel_shape = IR_node.get_attr('kernel_shape')

        if len(kernel_shape) == 4:
            height, width, input_channels, output_channels = kernel_shape
        elif len(kernel_shape) == 5:
            depth, height, width, input_channels, output_channels = kernel_shape
        else:
            raise NotImplementedError()

        output_shape = None

        # W should have shape (height, width, kernel_channels, output_channels), where kernel_channel = input_channels / groups
        W = self.weights_dict[IR_node.name]['weights']
        b = self.weights_dict[IR_node.name]['bias'] if has_bias else None


        stride_height, stride_width = IR_node.get_attr('strides')[1], IR_node.get_attr('strides')[2]

        # Dilations
        dilations = IR_node.get_attr('dilations', [1, 1])
        if is_deconv and not dilations == [1, 1]:
            raise ValueError("Unsupported non-unity dilation for Deconvolution layer")

        groups = IR_node.get_attr('group', 1)

        kernel_channels = input_channels // groups
        padding = self._get_padding(IR_node)

        if isinstance(padding, list):
            border_mode = "valid"
            # see protobuf
            padding_top, padding_left, padding_bottom, padding_right = padding[1], padding [2], padding[5], padding [6]
        else:
            border_mode = "same"
            padding_top, padding_left, padding_bottom, padding_right = 0, 0, 0, 0


        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name

        self.builder.add_convolution(name=IR_node.real_name,
                                     kernel_channels=kernel_channels,
                                     output_channels=output_channels,
                                     height=height,
                                     width=width,
                                     stride_height=stride_height,
                                     stride_width=stride_width,
                                     border_mode= border_mode,
                                     groups=groups,
                                     W=W,
                                     b=b,
                                     has_bias=has_bias,
                                     is_deconv=is_deconv,
                                     output_shape=output_shape,
                                     input_name=input_name,
                                     padding_top= padding_top,
                                     padding_left= padding_left,
                                     padding_bottom= padding_bottom,
                                     padding_right= padding_right,
                                     output_name=IR_node.real_name,
                                     dilation_factors=dilations)




    def emit_ConvTranspose(self, IR_node):
        """
        Convert convolution layer to coreml.
        """

        # assert False
        has_bias = IR_node.get_attr('use_bias', False)
        is_deconv = True

        # Get the weights.

        kernel_shape = IR_node.get_attr('kernel_shape')

        if len(kernel_shape) == 4:
            height, width, output_channels, kernel_channels = kernel_shape
            W = self.weights_dict[IR_node.name]['weights']
            W = W.reshape(kernel_shape)
            W = W.transpose((0, 1, 3, 2))
        elif len(kernel_shape) == 5:
            depth, height, width, output_channels, kernel_channels = kernel_shape
            W = self.weights_dict[IR_node.name]['weights']
            W = W.reshape(kernel_shape)
            W = W.transpose((0, 1, 2, 4, 3))
        else:
            raise NotImplementedError()


        output_shape = None
        b = self.weights_dict[IR_node.name]['bias'] if has_bias else None

        stride_height, stride_width = IR_node.get_attr('strides')[1], IR_node.get_attr('strides')[2]

        # Dilations
        dilations = IR_node.get_attr('dilations', [1, 1])
        if is_deconv and not dilations == [1, 1]:
            raise ValueError("Unsupported non-unity dilation for Deconvolution layer")

        groups = IR_node.get_attr('group', 1)

        padding = self._get_padding(IR_node)

        if isinstance(padding, list):
            border_mode = "valid"
            # see protobuf
            padding_top, padding_left, padding_bottom, padding_right = padding[1], padding [2], padding[5], padding [6]
        else:
            border_mode = "same"
            padding_top, padding_left, padding_bottom, padding_right = 0, 0, 0, 0


        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name

        self.builder.add_convolution(name=IR_node.real_name,
                                     kernel_channels=kernel_channels,
                                     output_channels=output_channels,
                                     height=height,
                                     width=width,
                                     stride_height=stride_height,
                                     stride_width=stride_width,
                                     border_mode= border_mode,
                                     groups=groups,
                                     W=W,
                                     b=b,
                                     has_bias=has_bias,
                                     is_deconv=is_deconv,
                                     output_shape=output_shape,
                                     input_name=input_name,
                                     padding_top= padding_top,
                                     padding_left= padding_left,
                                     padding_bottom= padding_bottom,
                                     padding_right= padding_right,
                                     output_name=IR_node.real_name,
                                     dilation_factors=dilations)



    def emit_DepthwiseConv(self, IR_node):
        # depth-wise convolution

        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name
        kernel_channels = 1
        is_deconv = False
        has_bias = IR_node.get_attr('use_bias', False)

        depth_multiplier = IR_node.get_attr('kernel_shape')[-1]

        W = self.weights_dict[IR_node.name]['weights']
        height, width, channels, n_filters = W.shape
        output_shape = None
        W = np.reshape(W,(height, width,1,channels * depth_multiplier))
        b = self.weights_dict[IR_node.name]['bias'] if has_bias else None

        # Dilations
        dilations = IR_node.get_attr('dilations', [1, 1])


        padding = self._get_padding(IR_node)



        if isinstance(padding, list):
            border_mode = "valid"
            # see protobuf
            padding_top, padding_left, padding_bottom, padding_right = padding[1], padding [2], padding[5], padding [6]
        else:
            border_mode = "same"
            padding_top, padding_left, padding_bottom, padding_right = 0, 0, 0, 0





        output_channels = W.shape[-1]
        groups = W.shape[-1]
        stride_height, stride_width = IR_node.get_attr('strides')[1], IR_node.get_attr('strides')[2]

        self.builder.add_convolution(name=IR_node.real_name,
                                     kernel_channels=kernel_channels,
                                     output_channels=output_channels,
                                     height=height,
                                     width=width,
                                     stride_height=stride_height,
                                     stride_width=stride_width,
                                     border_mode=border_mode,
                                     groups=groups,
                                     W=W,
                                     b=b,
                                     has_bias=has_bias,
                                     is_deconv=is_deconv,
                                     output_shape=output_shape,
                                     padding_top= padding_top,
                                     padding_left= padding_left,
                                     padding_bottom= padding_bottom,
                                     padding_right= padding_right,
                                     input_name=input_name,
                                     output_name=IR_node.real_name,
                                     dilation_factors=dilations)


    def emit_Pool(self, IR_node):
        """
        Convert pooling layer to coreml.
        """
        # Get input and output names
        input_name = self.IR_graph.get_node(IR_node.in_edges[0]).real_name

        # Pooling layer type
        pooling_type = IR_node.get_attr('pooling_type')
        if pooling_type == 'MAX':
            layer_type_str = 'MAX'
        elif pooling_type == 'AVG':
            layer_type_str = 'AVERAGE'
        else:
            raise TypeError("Pooling type %s not supported" % pooling_type)


        # if it's global, set the global flag
        global_pooling = IR_node.get_attr('global_pooling', False)
        dim = len(IR_node.get_attr('strides')) - 2


        if global_pooling:
            if dim == 2:



                stride_height, stride_width = tuple(IR_node.get_attr('strides')[1:-1])
                height, width = 1, 1

                # TODO  global pooling modification

                # Padding
                padding = self._get_padding(IR_node)

                if isinstance(padding, list):
                    padding_type = "VALID"
                    # see protobuf
                    padding_top, padding_left, padding_bottom, padding_right = padding[1], padding[2], padding[5], padding[6]
                else:
                    padding_type = "SAME"
                    padding_top, padding_left, padding_bottom, padding_right = 0, 0, 0, 0


            elif dim == 1:
                raise NotImplementedError()
                global_pooling = False
                _, width, channels = keras_layer.input_shape
                height = 1
                stride_height, stride_width = height, width
                padding_type = 'VALID'
            else:
                raise NotImplementedError()

        else:

            height, width = tuple(IR_node.get_attr('kernel_shape')[1:-1])
            stride_height, stride_width = tuple(IR_node.get_attr('strides')[1:-1])

            # Padding
            padding = self._get_padding(IR_node)
            if isinstance(padding, list):

                padding_type = "VALID"
                # see protobuf
                padding_top, padding_left, padding_bottom, padding_right = padding[1], padding [2], padding[5], padding [6]
            else:
                padding_type = "SAME"
                padding_top, padding_left, padding_bottom, padding_right = 0, 0, 0, 0


        self.builder.add_pooling(name=IR_node.name,
                                    height=height,
                                    width=width,
                                    stride_height=stride_height,
                                    stride_width=stride_width,
                                    layer_type=layer_type_str,
                                    padding_type=padding_type,
                                    padding_top= padding_top,
                                    padding_left= padding_left,
                                    padding_bottom= padding_bottom,
                                    padding_right= padding_right,
                                    input_name=input_name,
                                    output_name=IR_node.name,
                                    exclude_pad_area=True,
                                    is_global=global_pooling)


    def emit_Scale(self, IR_node):
        # Get input and output names
        input_name = self.IR_graph.get_node(IR_node.in_edges[0]).real_name

        weights = IR_node.get_attr('scale', False)
        weights = self.weights_dict[IR_node.name]['scale']
        has_bias = IR_node.get_attr('use_bias', False)
        if has_bias:
            bias = self.weights_dict[IR_node.name]['bias']


        shape_scale = self.weights_dict[IR_node.name]['shapeScale']
        if has_bias:
            shape_bias = self.weights_dict[IR_node.name]['shapeBias']



        self.builder.add_scale(name = IR_node.real_name,
                                        W = weights,
                                        b = bias,
                                        has_bias = has_bias,
                                        input_name = input_name,
                                        output_name =IR_node.name,
                                        shape_scale= [shape_scale],
                                        shape_bias= [shape_bias])



    def emit_UNKNOWN(self, IR_node):
        print(IR_node.name)


    def emit_Crop(self, IR_node):
        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name
        output_name=IR_node.real_name

        is_1d = False
        border = IR_node.get_attr('border')

        if is_1d:
            raise ValueError("Unrecognized padding option: %s" % (str(border)))
        else:
            if type(border) is int:
                top = left = bottom = right = border
            elif type(border) is list:
                # type: "list(int). A 1-D values of (leftBorder, topBorder, rightBorder, bottomBorder)."
                # This is central crop
                top, left = border[1], border[0]
                bottom, right = border[1], border[0]
            else:
                raise ValueError("Unrecognized padding option: %s" % (str(border)))

        # Now add the layer
        self.builder.add_crop(name = IR_node.name,
            left = left, right=right, top=top, bottom=bottom, offset = [0,0],
            input_names = [input_name], output_name=output_name
            )


    def emit_ReduceMean(self, IR_node):
        """
        Convert ReduceMean layer to coreml.
        """

        axis = IR_node.get_attr('axes', [1,2])

#       Allowed values: 'CHW', 'HW', 'C', 'H', 'W'
        if len(axis) == 1:
            if axis[0] == 0:
                axis_str = 'C'
            elif axis[0] == 1:
                axis_str = 'H'
            elif axis[0] == 2:
                axis_str = 'W'
        elif len(axis) == 2:
            axis_str = 'HW'
        elif len(axis) == 3:
            axis_str = 'CHW'

        # Get input and output names
        input_name = self.IR_graph.get_node(IR_node.in_edges[0]).real_name


        self.builder.add_reduce(IR_node.name,
                            input_name = input_name,
                            output_name = IR_node.name,
                            axis = axis_str,
                            mode = 'avg',
                            epsilon = 1e-6)


    def emit_DataInput(self, IR_node):
        """ Layers that can be skipped. """
        return


    def emit_Dropout(self, IR_node):
        """ Layers that can be skipped (because they are train time only. """
        IR_node.real_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name


    def emit_FullyConnected(self, IR_node):
        """
        Convert a dense layer to coreml.
        """
        # Get input and output names
        input_name = self.IR_graph.get_node(IR_node.in_edges[0]).real_name
        output_name = IR_node.out_edges[0]

        has_bias = IR_node.get_attr('use_bias')

        # Get the weights from keras
        W = self.weights_dict[IR_node.name]['weights'].T
        Wb = self.weights_dict[IR_node.name]['bias'].T if has_bias else None
        output_channels, input_channels = W.shape

        self.builder.add_inner_product(name=IR_node.name,
                                       W=W,
                                       b=Wb,
                                       input_channels=input_channels,
                                       output_channels=output_channels,
                                       has_bias=has_bias,
                                       input_name=input_name,
                                       output_name=IR_node.name)


    def emit_Flatten(self, IR_node):
        """
        Convert a flatten layer from keras to coreml.
        """
        # Get input and output names
        input_name = self.IR_graph.get_node(IR_node.in_edges[0]).real_name
        output_name = IR_node.out_edges[0]

        """
        # blob_order == 0 if the input blob needs not be rearranged
        # blob_order == 1 if the input blob needs to be rearranged
        blob_order = 0

        # using keras_layer.input.shape have a "?" (Dimension[None] at the front),
        # making a 3D tensor with unknown batch size 4D
        if len(keras_layer.input.shape) == 4:
            blob_order = 1
        """

        self.builder.add_flatten(name=IR_node.name, mode=1,
                                 input_name=input_name, output_name=IR_node.name)


    def emit_Reshape(self, IR_node):
        def ShapetrToTuple(string, batch_none = False):
            if batch_none == True:
                ls = [int(item) for item in string.split(', ')]
                ls.insert(0,None)
                return tuple(ls)
            else:
                ls = [int(item) for item in string.split(', ')]
                return tuple(ls)

        last_node = self.IR_graph.get_node(IR_node.in_edges[0]).layer
        input_shape_dims = last_node.attr["_output_shapes"].list.shape
        target_shape_dims = IR_node.IR_layer.attr["_output_shapes"].list.shape

        input_shape = ShapetrToTuple(IRGraph.shapeToStr(input_shape_dims[0]),True)
        target_shape = ShapetrToTuple(IRGraph.shapeToStr(target_shape_dims[0]))

        def get_coreml_target_shape(target_shape):
            if len(target_shape) == 1: #(D,)
                coreml_shape = (1,target_shape[0],1,1)
            elif len(target_shape) == 2: #(S,D)
                coreml_shape = target_shape + (1,1)
            elif len(target_shape) == 3: #(H,W,C)
                coreml_shape = (1, target_shape[2], target_shape[0], target_shape[1])
            else:
                coreml_shape = None
            return coreml_shape

        def get_mode(input_shape, target_shape):
            in_shape = input_shape[1:]
            if len(in_shape) == 3 or len(target_shape) == 3:
                    return 1
            else:
                return 0
        input_name = self.IR_graph.get_node(IR_node.in_edges[0]).real_name
        new_shape = get_coreml_target_shape(target_shape)
        mode = get_mode(input_shape, target_shape)

        self.builder.add_reshape(
            name=IR_node.real_name,
            input_name=input_name,
            output_name=IR_node.real_name,
            target_shape=new_shape,
            mode=mode)



    def _emit_activation(self, IR_node, act, params=None):
        # Get input and output names
        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name
        output_name = IR_node.real_name
        if not isinstance(params, list):
            params = [params]

        self.builder.add_activation(name=IR_node.real_name,
            non_linearity=act,
            input_name=input_name,
            output_name=output_name,
            params=params)


    # activation emit
    def emit_Relu(self, IR_node):
        self._emit_activation(IR_node, 'RELU')
    def emit_Tanh(self, IR_node):
        self._emit_activation(IR_node, 'TANH')
    def emit_PRelu(self, IR_node):
        self._emit_activation(IR_node, 'PRELU', IR_node.get_attr('gamma', 0) )
    def emit_LeakyRelu(self, IR_node):
        self._emit_activation(IR_node, 'LEAKYRELU', IR_node.get_attr('alpha', 0) )
    def emit_Elu(self,IR_node):
        self._emit_activation(IR_node, 'ELU',  IR_node.get_attr('alpha', 0)  )
    def emit_ThresholdedRelu(self, IR_node):
        self._emit_activation(IR_node, 'THRESHOLDEDRELU', IR_node.get_attr('alpha', 0) )
    def emit_ScaledTanh(self, IR_node):
        self._emit_activation(IR_node, 'SCALED_TANH', [IR_node.get_attr('alpha', 0),IR_node.get_attr('beta', 0)])
    def emit_linear(self, IR_node):
        self._emit_activation(IR_node, 'LINEAR', [IR_node.get_attr('alpha', 0),IR_node.get_attr('beta', 0)])
    def emit_SigmoidHard(self, IR_node):
        self._emit_activation(IR_node, 'SIGMOID_HARD', [IR_node.get_attr('alpha', 0),IR_node.get_attr('beta', 0)])
    def emit_ParametricSoftplus(self, IR_node):
        self._emit_activation(IR_node, 'PARAMETRICSOFTPLUS', [ IR_node.get_attr('alpha', 0),IR_node.get_attr('beta', 0) ])




    def emit_Softmax(self, IR_node):
        # Get input and output names
        input_name = self.IR_graph.get_node(IR_node.in_edges[0]).real_name
        output_name = IR_node.out_edges[0]
        self.builder.add_softmax(name=IR_node.name, input_name=input_name,
                                 output_name=IR_node.name)


    def emit_Sigmoid(self, IR_node):
        assert False
        code = "{:<15} = Activation(name = '{}', activation = 'sigmoid')({})".format(
                IR_node.replace_scope(IR_node.name),
                IR_node.name,
                IR_node.replace_scope(IR_node.in_edges[0]))
        return code


    def emit_Relu6(self, IR_node):

        layer = IR_node.real_name
        input_name, output_name = (IR_node.IR_layer.input[0], IR_node.IR_layer.name)

        relu_output_name = output_name + '_relu'
        self.builder.add_activation(layer, 'RELU', input_name, relu_output_name)
        # negate it
        neg_output_name = relu_output_name + '_neg'
        self.builder.add_activation(layer+'__neg__', 'LINEAR', relu_output_name,
                neg_output_name,[-1.0, 0])
        # apply threshold
        clip_output_name = relu_output_name + '_clip'
        self.builder.add_unary(layer+'__clip__', neg_output_name, clip_output_name,
                'threshold', alpha = -6.0)
        # negate it back
        self.builder.add_activation(
            layer + '_neg2',
            'LINEAR',
            clip_output_name,
            output_name,
            [-1.0, 0])

    def emit_Gather(self, IR_node):
        raise NotImplementedError()
        W = self.weights_dict[IR_node.name]['weights']
        if W.ndim == 2:
            vocab_size = W.shape[0]
            output_channels = W.shape[1]
            builder.add_embedding(
                name=IR_node.real_name,
                W = W,
                b = None,
                input_dim = vocab_size,
                output_channels = output_channels,
                has_bias=False,
                input_name=input_name,
                output_name=IR_node.real_name)
        else:
            raise NotImplementedError()

    def emit_RNNs(self, IR_node, func):
        assert False
        # for Keras
        if "dropout" in IR_node.IR_layer.attr:
            dropout_str = ",dropout = {}, recurrent_dropout = {}".format(
                    IR_node.IR_layer.attr['dropout'].f,
                    IR_node.IR_layer.attr['recurrent_dropout'].f)
        else:
            dropout_str = ""

        code = "{:<15} = {}(units = {}, use_bias = {} {})({})".format(
                IR_node.name,
                func,
                IR_node.IR_layer.attr['units'].i,
                IR_node.IR_layer.attr['use_bias'].b,
                dropout_str,
                IR_node.in_edges[0])

        return code


    def emit_LSTM(self, IR_node):
        return self.emit_RNNs(IR_node, "LSTM")


    def emit_GRU(self, IR_node):
        return self.emit_RNNs(IR_node, "GRU")


    def emit_Add(self, IR_node):
        self._emit_merge(IR_node, 'ADD')


    def emit_Concat(self, IR_node):
        self._emit_merge(IR_node, "CONCAT")


    def emit_BatchNorm(self, IR_node):
        """
        Convert a Batch Normalization layer.
        """
        # Get input and output names
        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name


        axis = IR_node.get_attr('axis', -1)
        nb_channels = IR_node.get_attr('_output_shapes')[0].dim[axis].size


        # Set parameters
        # Parameter arrangement in Keras: gamma, beta, mean, variance
        weights = self.weights_dict[IR_node.name]
        mean = weights['mean']
        std = weights['var']
        gamma = weights.get('scale', np.ones(mean.shape))
        beta = weights.get('bias', np.zeros(mean.shape))

        # compute adjusted parameters
        # Reference: parameter transformation https://github.com/apple/coremltools/issues/153
        variance = std * std
        f = 1.0 / np.sqrt(std + IR_node.get_attr('epsilon'))
        gamma1 = gamma*f
        beta1 = beta - gamma*mean*f
        mean[:] = 0.0 #mean
        variance[:] = 1.0 - .00001 #stddev


        self.builder.add_batchnorm(
            name=IR_node.real_name,
            channels = nb_channels,
            gamma = gamma1,
            beta = beta1,
            mean = mean,
            variance = variance,
            input_name = input_name,
            output_name=IR_node.real_name)



    def emit_Pad(self, IR_node):
        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name
        output_name=IR_node.real_name
        is_1d = False
        padding = IR_node.get_attr('pads')
        if is_1d:
            raise ValueError("Unrecognized padding option: %s" % (str(padding)))
        else:
            if type(padding) is int:
                top = left = bottom = right = padding
            elif type(padding) is list:
                top, left = padding[1], padding [2]
                bottom, right = padding[5], padding [6]
            else:
                raise ValueError("Unrecognized padding option: %s" % (str(padding)))

        # padding type TODO
        # Type of the padding. Can be one of 'constant', 'reflection' or 'replication
        padding_type = IR_node.get_attr('mode', 'CONSTANT')
        if padding_type == 'CONSTANT':
            padding_type = 'constant'
        elif padding_type == 'REFLECT':
            padding_type = 'reflection'
        elif padding_type == 'SYMMETRIC':
            padding_type = 'replication'


        # Now add the layer
        self.builder.add_padding(name = IR_node.name,
            left = left, right=right, top=top, bottom=bottom, value = 0,
            input_name = input_name, output_name=output_name, padding_type = padding_type
            )


    def emit_Squeeze(self, IR_node):
        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name
        output_name=IR_node.real_name

        self.builder.add_bias(name = IR_node.name,
                              b = 0,
                              input_name = input_name,
                              output_name = output_name,
                              shape_bias = [1])
        # self.emit_Flatten(IR_node)


    def emit_LRN(self, IR_node):
        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name
        output_name = IR_node.real_name
        alpha = IR_node.get_attr('alpha')
        beta = IR_node.get_attr('beta')
        k = IR_node.get_attr('k')
        depth_radius = int(IR_node.get_attr('size'))
        #  depth_radius: Half-width of the 1-D normalization window."
        self.builder.add_lrn(output_name, input_name, output_name,
                          alpha=alpha,
                          beta=beta,
                          local_size=2*depth_radius-1,
                          k=k)


    def emit_SeparableConv(self, IR_node):

        input_name = self.IR_graph.get_parent(IR_node.name, [0]).real_name
        output_name = IR_node.real_name



        strides = IR_node.get_attr('strides')
        stride_height, stride_width = (strides[1], strides[2])

        # Get the weights
        W0 = self.weights_dict[IR_node.name]['depthwise_filter']
        W1 = self.weights_dict[IR_node.name]['pointwise_filter']

        padding = IR_node.get_attr('auto_pad').split('_')[0].lower()
        has_bias = IR_node.get_attr('use_bias')
        b = self.weights_dict[IR_node.name]['bias'] if has_bias else None

        output_blob_shape = IR_node.get_attr('_output_shapes')
        shape = shape_to_list(output_blob_shape[0])
        output_channels = shape[-1]

        height, width, input_channels, depth_mult = W0.shape

        W0 = np.reshape(W0, (height, width, 1, input_channels * depth_mult))

        intermediate_name = input_name + '_intermin_'

        self.builder.add_convolution(name = IR_node.name + '_step_1',
             kernel_channels = 1,
             output_channels = input_channels * depth_mult,
             height = height,
             width = width,
             stride_height = stride_height,
             stride_width = stride_width,
             border_mode = padding,
             groups = input_channels,
             W = W0,
             b = None,
             has_bias = False,
             is_deconv = False,
             output_shape = None,
             input_name = input_name,
             output_name = intermediate_name,
             dilation_factors = [1,1])

        self.builder.add_convolution(name = IR_node.name + '_step_2',
                kernel_channels = input_channels * depth_mult,
                output_channels = output_channels,
                height = 1,
                width = 1,
                stride_height = 1,
                stride_width = 1,
                border_mode = padding,
                groups = 1,
                W = W1,
                b = b,
                has_bias = has_bias,
                is_deconv = False,
                output_shape = None,
                input_name = intermediate_name,
                output_name = output_name,
                dilation_factors = [1,1])


    def emit_Slice(self, IR_node):
        pass
    def emit_Const(self, IR_node):
        pass

    def emit_Shape(self, IR_node):
        pass
    def emit_Pack(self, IR_node):
        pass



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import os
from six.moves import xrange

from mmdnn.conversion.common.IR.IR_graph import IRGraph, IRGraphNode
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.DataStructure.emitter import Emitter
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.keras.extra_layers import Scale
from mmdnn.conversion.rewriter.folder import Folder


class Keras2Emitter(Emitter):

    dtype_map = {
        graph_pb2.DT_FLOAT16 : "float16",
        graph_pb2.DT_FLOAT32 : "float32",
        graph_pb2.DT_FLOAT64 : "float64",
        graph_pb2.DT_INT16 : "int16",
        graph_pb2.DT_INT32 : "int32",
        graph_pb2.DT_INT64 : "int64",
        graph_pb2.DT_UINT8 : "uint8",
        graph_pb2.DT_UINT16 : "uint16"
    }


    def __init__(self, model):
        super(Keras2Emitter, self).__init__()
        from six import string_types as _string_types
        if isinstance(model, _string_types):
            network_path = model
        else:
            network_path = model[0]
            weight_path = model[1]
            self._load_weights(weight_path)

        self.IR_graph = IRGraph(network_path)
        self.IR_graph.build()
        self.yolo_parameter = []
        self.region_parameter = []
        self.layers_codes_count = dict()

        folder = Folder(self.IR_graph, self.weights_dict)
        folder.fold()


    @property
    def header_code(self):
        return """import keras
from keras.models import Model
from keras import layers
import keras.backend as K
import numpy as np
from keras.layers.core import Lambda
import tensorflow as tf


weights_dict = dict()
def load_weights_from_file(weight_file):
    try:
        weights_dict = np.load(weight_file).item()
    except:
        weights_dict = np.load(weight_file, encoding='bytes').item()

    return weights_dict


def set_layer_weights(model, weights_dict):
    for layer in model.layers:
        if layer.name in weights_dict:
            cur_dict = weights_dict[layer.name]
            current_layer_parameters = list()
            if layer.__class__.__name__ == "BatchNormalization":
                if 'scale' in cur_dict:
                    current_layer_parameters.append(cur_dict['scale'])
                if 'bias' in cur_dict:
                    current_layer_parameters.append(cur_dict['bias'])
                current_layer_parameters.extend([cur_dict['mean'], cur_dict['var']])
            elif layer.__class__.__name__ == "Scale":
                if 'scale' in cur_dict:
                    current_layer_parameters.append(cur_dict['scale'])
                if 'bias' in cur_dict:
                    current_layer_parameters.append(cur_dict['bias'])
            elif layer.__class__.__name__ == "SeparableConv2D":
                current_layer_parameters = [cur_dict['depthwise_filter'], cur_dict['pointwise_filter']]
                if 'bias' in cur_dict:
                    current_layer_parameters.append(cur_dict['bias'])
            elif layer.__class__.__name__ == "Embedding":
                current_layer_parameters.append(cur_dict['weights'])
            else:
                # rot weights
                current_layer_parameters = [cur_dict['weights']]
                if 'bias' in cur_dict:
                    current_layer_parameters.append(cur_dict['bias'])
            model.get_layer(layer.name).set_weights(current_layer_parameters)

    return model


def KitModel(weight_file = None):
    global weights_dict
    weights_dict = load_weights_from_file(weight_file) if not weight_file == None else None
        """


    def gen_code(self, phase):
        self.add_body(0, self.header_code)
        for layer in self.IR_graph.topological_sort:
            current_node = self.IR_graph.get_node(layer)
            node_type = current_node.type

            if hasattr(self, "emit_" + node_type):
                # print("Converting layer {}({})".format(current_node.name, node_type))
                func = getattr(self, "emit_" + node_type)
                line = func(current_node)
                if line:
                    self.add_body(1, line)
            else:
                print("KerasEmitter has not supported operator [%s]." % (node_type))
                self.emit_UNKNOWN(current_node)

        self.add_body(1, "{:<15} = Model(inputs = [{}], outputs = [{}])".format(
            "model",
            ', '.join([self.IR_graph.get_node(name).real_variable_name for name in self.IR_graph.input_layers if self.IR_graph.get_node(name).type != 'Const']),
            ', '.join([self.IR_graph.get_node(name).real_variable_name for name in self.IR_graph.output_layers if self.IR_graph.get_node(name).type != 'Pack'])))
        self.add_body(1, ["set_layer_weights(model, weights_dict)", "return model"])

        for i in self.used_layers:
            func = getattr(self, "_layer_" + i)
            func()

        self.add_body(0, "")
        for code in self.layers_codes.values():
            self.add_body(0, code)

        return self.body_code


    @staticmethod
    def shapeToStr(shapes):
        return ', '.join('%s' % i for i in filter(lambda x:x > 0, shapes))


    def _emit_activation(self, IR_node, op, in_scope=False):
        if in_scope:
            code = "{:<15} = keras.activations.get('{}')({})".format(
                IR_node.variable_name,
                op,
                self.parent_variable_name(IR_node))
        else:
            code =  "{:<15} = layers.Activation(name='{}', activation='{}')({})".format(
                IR_node.variable_name,
                IR_node.name,
                op,
                self.parent_variable_name(IR_node))

        return code


    def _emit_merge(self, IR_node, func):
        if len(IR_node.in_edges) == 1:
            IR_node.in_edges.append(IR_node.in_edges[0])
        inputs = ', '.join('%s' % self.parent_variable_name(IR_node, i) for i in IR_node.in_edges)
        axis = ' axis = {},'.format(IR_node.get_attr('axis')) if 'axis' in IR_node.layer.attr else ""
        code =  "{:<15} = layers.{}(name = '{}', inputs = [{}])".format(
            IR_node.variable_name,
            func,
            IR_node.name,
            inputs)
        return code


    @staticmethod
    def _convert_padding(padding):
        padding = convert_onnx_pad_to_tf(padding)[1:-1]

        for idx, pad in enumerate(padding):
            padding[idx] = tuple(pad)
        padding = tuple(padding)
        return padding


    def _defuse_padding(self, IR_node, in_scope=False):
        auto_pad = IR_node.get_attr('auto_pad')

        if auto_pad != None and auto_pad.startswith("SAME"):
            input_node = self.parent_variable_name(IR_node)
            padding = 'same'
            return input_node, padding
        else:

            padding = IR_node.get_attr("pads")

            if padding != None:
                padding = self._convert_padding(padding)

                if is_valid_padding(padding) == False:
                    input_node = IR_node.variable_name + '_input'
                    self.add_body(1, "{:<15} = layers.ZeroPadding{}D(padding = {})({})".format(
                        input_node,
                        len(padding),
                        padding,
                        self.parent_variable_name(IR_node)))
                else:
                    input_node = self.parent_variable_name(IR_node)
            else:
                input_node = self.parent_variable_name(IR_node)


            # TODO
            return input_node, 'valid'
            # return input_node, 'same'


    def _emit_convolution(self, IR_node, conv_type):
        self.used_layers.add('Conv')
        # assert IR_node.get_attr('group', 1) == 1
        group = IR_node.get_attr("group", 1)

        if conv_type.endswith('Transpose'):
            filters = IR_node.get_attr('kernel_shape')[-2]
        else:
            filters = IR_node.get_attr('kernel_shape')[-1]

        filters_str = 'filters={}'.format(filters) if not conv_type.endswith('DepthwiseConv2D') else 'depth_multiplier={}'.format(filters)
        # change dw from filters to 1


        input_node, padding = self._defuse_padding(IR_node)

        dilations = IR_node.get_attr('dilations')

        if not dilations or len(dilations) == 2:
            # reset the default dilation
            dilations = [1] * len(IR_node.get_attr('kernel_shape'))

        code = "{:<15} = convolution(weights_dict, name='{}', input={}, group={}, conv_type='{}', {}, kernel_size={}, strides={}, dilation_rate={}, padding='{}', use_bias={})".format(
            IR_node.variable_name,
            IR_node.name,
            input_node,
            group,
            conv_type,
            filters_str,
            tuple(IR_node.get_attr('kernel_shape')[:-2]),
            tuple(IR_node.get_attr('strides')[1:-1]),
            tuple(dilations[1:-1]),
            padding,
            IR_node.get_attr('use_bias'))

        return code


    def emit_ConvTranspose(self, IR_node, in_scope=False):
        dim = len(IR_node.get_attr('kernel_shape')) - 2
        return self._emit_convolution(IR_node, 'layers.Conv{}DTranspose'.format(dim))


    def emit_Conv(self, IR_node, in_scope=False):
        dim = len(IR_node.get_attr('kernel_shape')) - 2
        return self._emit_convolution(IR_node, 'layers.Conv{}D'.format(dim))


    #############
    # Operators #
    #############

    def emit_UNKNOWN(self, IR_node, in_scope=False):
        print (IR_node.name)


    def emit_Mul(self, IR_node, in_scope=False):

        if in_scope:
            code = "{:<15} = {} * {}".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node, [1]))
            return code

        node_1 = self.IR_graph.get_node(IR_node.in_edges[0])
        node_2 = self.IR_graph.get_node(IR_node.in_edges[1])

        if node_1.type == 'Constant' or node_2.type == 'Constant':
            self.used_layers.add('Mul_Constant')
            if node_1.type == 'Constant': 
                weight_factor = node_1.get_attr('value')
                code = "{:<15} = mul_constant(weight_factor={}, layer_name= {})".format(
                    IR_node.variable_name,
                    weight_factor,
                    self.parent_variable_name(IR_node, [1]))
            else: 
                weight_factor = node_2.get_attr('value')
                code = "{:<15} = mul_constant(weight_factor={}, layer_name= {})".format(
                    IR_node.variable_name,
                    weight_factor,
                    self.parent_variable_name(IR_node))
        else:
            self.used_layers.add('Mul')
            code = "{:<15} = my_mul(name='{}')([{}, {}])".format(
                IR_node.variable_name,
                IR_node.name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node, [1]))  
        return code


    def emit_Sub(self, IR_node, in_scope=False):
        if in_scope:
            code = "{:<15} = {} - {}".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node, [1]))
            return code

        self.used_layers.add('Sub')
        code = "{:<15} = my_sub()({}, {})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]))

        # code = self._emit_merge(IR_node, "subtract")
        return code


    def emit_Add(self, IR_node, in_scope=False):
        if in_scope:
            code = "{:<15} = {} + {}".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node, [1]))
            return code

        self.used_layers.add('Add')
        code = "{:<15} = my_add()([{}, {}])".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]))
        return code


    def emit_DataInput(self, IR_node, in_scope=False):
        shape_str = IRGraph.shapeToStr(IR_node.IR_layer.attr["shape"].shape)
        dtype_str = ", dtype = '{}'".format(self.dtype_map[IR_node.layer.attr['dtype'].type]) if 'dtype' in IR_node.layer.attr else ""

        code = "{:<15} = layers.Input(name = '{}', shape = ({},) {})".format(
            IR_node.variable_name,
            IR_node.name,
            shape_str,
            dtype_str)
        return code


    def emit_Dropout(self, IR_node, in_scope=False):
        seed = 'None'
        if 'seed' in IR_node.IR_layer.attr:
            seed = IR_node.IR_layer.attr['seed'].i

        code = "{:<15} = layers.Dropout(name = '{}', rate = {}, seed = {})({})".format(
            IR_node.variable_name,
            IR_node.name,
            IR_node.IR_layer.attr["keep_prob"].f,
            seed,
            self.parent_variable_name(IR_node))
        return code


    def emit_FullyConnected(self, IR_node, in_scope=False):
        if in_scope:
            code = "{:<15} = K.bias_add(K.dot({}, K.variable(weights_dict['{}']['weights'])), K.variable(weights_dict['{}']['bias']))".format(
                IR_node.variable_name, 
                self.parent_variable_name(IR_node),
                IR_node.name,
                IR_node.name)
        else:
            code = "{:<15} = layers.Dense(name = '{}', units = {}, use_bias = {})({})".format(
                IR_node.variable_name,
                IR_node.name,
                IR_node.get_attr('units'),
                IR_node.get_attr('use_bias'),
                self.parent_variable_name(IR_node))
        return code


    def emit_Flatten(self, IR_node, in_scope=False):
        self.used_layers.add('Flatten')
        code = "{:<15} = __flatten(name = '{}', input = {})".format(
            IR_node.variable_name,
            IR_node.name,
            self.parent_variable_name(IR_node))
        return code


    def emit_Pool(self, IR_node, in_scope=False):
        codes = list()
        dim = len(IR_node.get_attr("strides")) - 2

        pooling_type = IR_node.get_attr('pooling_type')
        if  pooling_type == "MAX":
            pool_name = "MaxPooling{}D".format(dim)
        elif pooling_type == "AVG":
            pool_name = "AveragePooling{}D".format(dim)
        else:
            print(pooling_type)
            assert False

        # TODO
        if IR_node.layer.attr['global_pooling'].b:

            shape_str = IR_node.get_attr("shape_coreml")
            if shape_str:
                shape_str = ','.join([str(i) for i in shape_str])

                codes.append("{:<15} = layers.Global{}(name = '{}')({})".format(
                    IR_node.variable_name+'before',
                    pool_name,
                    IR_node.name,
                    self.parent_variable_name(IR_node)))

                #  when converting from coreml model, reshape is needed after the global pooling
                codes.append("{:<15} = layers.Reshape(name = '{}', target_shape = ({},))({})".format(
                    IR_node.variable_name,
                    IR_node.name + 'reshape',
                    shape_str,
                    IR_node.variable_name+'before'))
            else:
                codes.append("{:<15} = layers.Global{}(name = '{}')({})".format(
                IR_node.variable_name,
                pool_name,
                IR_node.name,
                self.parent_variable_name(IR_node)))

        else:
            dilations = IR_node.get_attr('dilations')
            if dilations:
                for e in IR_node.get_attr('dilations'):
                    assert e == 1

            pool_size = IR_node.get_attr('kernel_shape')[1:-1]

            strides = IR_node.get_attr('strides')[1:-1]
            padding = IR_node.get_attr('pads')[1:dim]

            if pooling_type == "AVG" and pool_size.count(pool_size[0]) == len(pool_size) and strides[0] == 1 and strides.count(strides[0]) == len(strides) and padding.count(padding[0]) == len(padding) and pool_size[0] == padding[0]*2 + 1:
                pool_size = ', '.join('%s' % i for i in pool_size)
                strides = ', '.join('%s' % i for i in strides)
                codes.append("{:<15} = layers.{}(name = '{}', pool_size = ({}), strides = ({}), padding = '{}')({})".format(
                    IR_node.variable_name,
                    pool_name,
                    IR_node.name,
                    pool_size,
                    strides,
                    'same',
                    self.parent_variable_name(IR_node)
                    ))


            else:

                pool_size = ', '.join('%s' % i for i in pool_size)
                strides = ', '.join('%s' % i for i in strides)
                input_node, padding = self._defuse_padding(IR_node)

                codes.append("{:<15} = layers.{}(name = '{}', pool_size = ({}), strides = ({}), padding = '{}')({})".format(
                    IR_node.variable_name,
                    pool_name,
                    IR_node.name,
                    pool_size,
                    strides,
                    padding,
                    input_node))
        return codes


    def emit_Reshape(self, IR_node, in_scope=False):
        shape_str = self.shapeToStr(IR_node.IR_layer.attr["shape"].list.i)
        code = "{:<15} = layers.Reshape(name = '{}', target_shape = ({},))({})".format(
            IR_node.variable_name,
            IR_node.name,
            shape_str,
            self.parent_variable_name(IR_node))
        return code


    def emit_Tanh(self, IR_node, in_scope=False):
        code = self._emit_activation(IR_node, 'tanh', in_scope)
        return code

    def emit_Relu(self, IR_node, in_scope=False):
        code = self._emit_activation(IR_node, 'relu', in_scope)
        return code

    def emit_Softmax(self, IR_node, in_scope=False):
        code = self._emit_activation(IR_node, 'softmax', in_scope)
        return code

    def emit_Sigmoid(self, IR_node, in_scope=False):
        code = self._emit_activation(IR_node, 'sigmoid', in_scope)
        return code

    def emit_Embedding(self, IR_node, in_scope=False):

        code = "{:<15} = layers.Embedding(name = '{}', input_dim = {}, output_dim = {}, mask_zero = {})({})".format(
            IR_node.variable_name,
            IR_node.name,
            IR_node.get_attr('input_dim'),
            IR_node.get_attr('output_dim'),
            IR_node.get_attr('mask_zero'),
            IR_node.in_edges[0])
        return code


    def emit_RNNs(self, IR_node, func):
        # for Keras
        if "dropout" in IR_node.IR_layer.attr:
            dropout_str = ",dropout = {}, recurrent_dropout = {}".format(
                    IR_node.IR_layer.attr['dropout'].f,
                    IR_node.IR_layer.attr['recurrent_dropout'].f)
        else:
            dropout_str = ""

        code = "{:<15} = layers.{}(units = {}, use_bias = {} {})({})".format(
                IR_node.name,
                func,
                IR_node.IR_layer.attr['units'].i,
                IR_node.IR_layer.attr['use_bias'].b,
                dropout_str,
                IR_node.in_edges[0])

        return code


    def emit_LSTM(self, IR_node, in_scope=False):
        return self.emit_RNNs(IR_node, "LSTM")


    def emit_GRU(self, IR_node, in_scope=False):
        return self.emit_RNNs(IR_node, "GRU")


    def emit_Concat(self, IR_node, in_scope=False):
        inputs = ', '.join('%s' % self.parent_variable_name(IR_node, s) for s in IR_node.in_edges)
        if in_scope:
            code = "{:<15} = K.concatenate([{}])".format(
                IR_node.variable_name,
                inputs)
        else:
            code = self._emit_merge(IR_node, "concatenate")
        return code


    def emit_BatchNorm(self, IR_node, in_scope=False):
        axis = IR_node.layer.attr['axis'].i if 'axis' in IR_node.layer.attr else -1

        code = "{:<15} = layers.BatchNormalization(name = '{}', axis = {}, epsilon = {}, center = {}, scale = {})({})".format(
            IR_node.variable_name,
            IR_node.name,
            axis,
            IR_node.layer.attr['epsilon'].f,
            IR_node.layer.attr['bias'].b,
            IR_node.layer.attr['scale'].b,
            self.parent_variable_name(IR_node))
        return code


    def emit_Scale(self, IR_node, in_scope=False):
        self.used_layers.add('Scale')
        axis = IR_node.layer.attr['axis'].i if 'axis' in IR_node.layer.attr else -1

        code = "{:<15} = Scale(name = '{}', axis = {}, center = {}, scale = {})({})".format(
            IR_node.variable_name,
            IR_node.name,
            axis,
            IR_node.layer.attr['use_bias'].b,
            True,
            self.parent_variable_name(IR_node))
        return code


    def emit_Pad(self, IR_node, in_scope=False):
        mode = IR_node.get_attr('mode', 'constant')
        mode = mode.lower()
        if mode == "constant":
            func = "ZeroPadding"
        else:
            raise NotImplementedError()

        dim = len(IR_node.get_attr('pads')) // 2 - 2

        padding = self._convert_padding(IR_node.get_attr('pads'))
        code = "{:<15} = layers.{}{}D(name='{}', padding={})({})".format(
            IR_node.variable_name,
            func,
            dim,
            IR_node.name,
            padding,
            self.parent_variable_name(IR_node))
        return code


    def emit_Squeeze(self, IR_node, in_scope=False):
        return self.emit_Flatten(IR_node)


    def emit_ReduceMean(self, IR_node, in_scope=False):
        axes = ', '.join('%s' % i for i in IR_node.get_attr('axes'))

        code = "{:<15} = layers.Lambda(lambda x: K.mean(x, axis=[{}], keepdims={}))({})".format(
            IR_node.variable_name,
            axes,
            IR_node.get_attr('keepdims'),
            self.parent_variable_name(IR_node))
        return code


    def emit_LRN(self, IR_node, in_scope=False):
        self.used_layers.add(IR_node.type)
        code = "{:<15} = LRN(size = {}, alpha = {}, beta = {}, k = {}, name = '{}')({})".format(
            IR_node.variable_name,
            IR_node.get_attr('size'),
            IR_node.get_attr('alpha'),
            IR_node.get_attr('beta'),
            IR_node.get_attr('k'),
            IR_node.name,
            self.parent_variable_name(IR_node))
        return code

    def emit_Split(self, IR_node, in_scope=False):
        if in_scope:
            axis = IR_node.get_attr('axis')
            split_num = IR_node.get_attr('split')
            segment_len = "K.int_shape({})[{}]//{}".format(self.parent_variable_name(IR_node),axis, split_num)
            split_str = '[' + ','.join(':' for i in range(axis)) + ',{}:{},...]'
            split_strs = []

            for i in range(split_num-1):
                split_strs.append(self.parent_variable_name(IR_node)+split_str.format(str(i)+'*'+ segment_len, str(i+1)+'*'+segment_len))

            split_strs.append(self.parent_variable_name(IR_node)+split_str.format(str(split_num-1)+'*'+segment_len, ''))

            code = "{:<15} = {}".format(IR_node.variable_name, ', '.join(split_strs))

        else:
            self.used_layers.add(IR_node.type)
            code = "{:<15} = __split(input={}, split_num={}, axis={})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                IR_node.get_attr('split'),
                IR_node.get_attr('axis'))

        return code
    
    def emit_Unsqueeze(self, IR_node, in_scope=False):
        self.used_layers.add(IR_node.type)

        code = "{:<15} = __unsqueeze(input={}, axis={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('axes')[0])
        return code

    def emit_Constant(self, IR_node, in_scope=False):

        if in_scope:
            if IR_node.get_attr('value'):
                code = "{:<15} = K.constant({})".format(IR_node.variable_name, IR_node.get_attr('value'))
            else:
                code = "{:<15} = K.constant(weights_dict['{}']['value'])".format(IR_node.variable_name, IR_node.name)
            return code
        else:
           pass 


    def emit_Shape(self, IR_node, in_scope=False):
        self.used_layers.add(IR_node.type)

        code = "{:<15} = __shape(input={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node))
        return code


    def emit_Fill(self, IR_node, in_scope=False):
        self.used_layers.add(IR_node.type)
        code = "{:<15} = __fill(input={}, value={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('value'))
        
        return code


    def emit_Slice(self, IR_node, in_scope=False):
        # It arouses some problems:
        # it can be implemented by Lambda Layer
        # https://github.com/keras-team/keras/issues/890

        self.used_layers.add(IR_node.type)

        extra_str = ""
        if IR_node.get_attr('strides'):
            extra_str += "strides={}".format(IR_node.get_attr('strides'))
        if IR_node.get_attr('begin_mask'):
            extra_str += ", begin_mask={}".format(IR_node.get_attr('begin_mask'))
        if IR_node.get_attr('end_mask'):
            extra_str += ", end_mask={}".format(IR_node.get_attr('end_mask'))
        if IR_node.get_attr('shrink_axis_mask'):
            extra_str += ", shrink_axis_mask={}".format(IR_node.get_attr('shrink_axis_mask'))

        code = "{:<15} = __slice({}, {}, {}, {})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('starts'),
            IR_node.get_attr('ends'),
            extra_str)
        return code

    def emit_Unstack(self, IR_node, in_scope=False):
        self.used_layers.add(IR_node.type)

        code = "{:<15} = __unstack(input={}, num={}, axis={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('num'),
            IR_node.get_attr('axis'))
        return code

    def emit_Pack(self, IR_node, in_scope=False):
        pass


    def emit_SeparableConv(self, IR_node, in_scope=False):
        assert len(IR_node.get_attr("strides")) == 4
        return self._emit_convolution(IR_node, "layers.SeparableConv2D")


    def emit_Relu6(self, IR_node, in_scope=False):
        try:
            # Keras == 2.1.6
            from keras.applications.mobilenet import relu6
            str_relu6 = 'keras.applications.mobilenet.relu6'
            code = "{:<15} = layers.Activation({}, name = '{}')({})".format(
                IR_node.variable_name,
                str_relu6,
                IR_node.name,
                self.IR_graph.get_node(IR_node.in_edges[0]).real_variable_name)
            return code

        except:
            # Keras == 2.2.2
            from keras.layers import ReLU
            code = "{:<15} = layers.ReLU(6, name = '{}')({})".format(
                IR_node.variable_name,
                IR_node.name,
                self.IR_graph.get_node(IR_node.in_edges[0]).real_variable_name)
            return code


    def emit_DepthwiseConv(self, IR_node, in_scope=False):
        try:
            from keras.applications.mobilenet import DepthwiseConv2D
            return self._emit_convolution(IR_node, 'keras.applications.mobilenet.DepthwiseConv2D')
        except:
            return self._emit_convolution(IR_node, 'layers.DepthwiseConv2D')


    def emit_Crop(self, IR_node, in_scope=False):
        border = IR_node.get_attr('border')
        rank = len(border) // 2
        cropping = []
        for idx in xrange(rank):
            cropping.append(tuple([border[idx * 2], border[idx * 2 + 1]]))

        code = "{:<15} = layers.Cropping{}D(cropping={}, name='{}')({})".format(
            IR_node.variable_name,
            rank,
            tuple(cropping),
            IR_node.name,
            self.parent_variable_name(IR_node))
        return code


    def emit_LeakyRelu(self, IR_node, in_scope=False):
        code = "{:<15} = layers.LeakyReLU(name='{}', alpha = {})({})".format(
            IR_node.variable_name,
            IR_node.name,
            IR_node.get_attr('alpha'),
            self.parent_variable_name(IR_node))
        return code

    def emit_UpSampling2D(self, IR_node, in_scope=False):
        code = "{:<15} = layers.UpSampling2D(name='{}', size= ({}), data_format = 'channels_last')({})".format(
            IR_node.variable_name,
            IR_node.name,
            IR_node.get_attr('scales'),
            self.parent_variable_name(IR_node))
        return code


    def emit_SpaceToDepth(self, IR_node, in_scope=False):
        self.used_layers.add(IR_node.type)
        assert IR_node.get_attr('blocksize') == 2
        # TODO: arguments won't be saved in keras export model

        blocksize = "arguments={'blocksize': %d}" % 2
        code = "{:<15} = layers.Lambda(space_to_depth, {}, name='{}')({})".format(
            IR_node.variable_name,
            blocksize,
            IR_node.name,
            self.parent_variable_name(IR_node))
        return code


    def emit_Maxmum(self, IR_node, in_scope=False):
        if in_scope:
            code = "{:<15} = K.maxmum({}, {})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node, [1])
            )
            return code
        else:
            return self._emit_merge(IR_node, 'Maxmum')


    def emit_Minimum(self, IR_node, in_scope=False):
        if in_scope:
            code = "{:<15} = K.minimum({}, {})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node, [1])
            )
            return code
        else:
            return self._emit_merge(IR_node, 'Minimum')


    def emit_yolo(self, IR_node, in_scope=False):
        self.used_layers.add('Yolo')
        self.yolo_parameter = [IR_node.get_attr('anchors'),
            IR_node.get_attr('classes'),
            IR_node.get_attr("ignore_thresh"),
            IR_node.get_attr("jitter")]
        code = "{:<15} = {}".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node))
        return code


    def emit_region(self, IR_node, in_scope=False):
        self.used_layers.add('Region')
        code = "{:<15} = {}".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node))
        self.region_parameter = [IR_node.get_attr('anchors'),
            IR_node.get_attr('classes'),
            IR_node.get_attr("thresh"),
            IR_node.get_attr("softmax"),
            IR_node.get_attr("bias_match"),
            IR_node.get_attr("jitter"),
            IR_node.get_attr("num"),
            IR_node.get_attr("random"),
            IR_node.get_attr("coords"),
            IR_node.get_attr("absolute"),
            IR_node.get_attr("rescore"),
            IR_node.get_attr("class_scale"),
            IR_node.get_attr("object_scale"),
            IR_node.get_attr("noobject_scale"),
            IR_node.get_attr("coord_scale"),
            ]
        return code

    def emit_Scope(self, IR_node, in_scope=False):
        if hasattr(self, '_emit_' + IR_node.pattern):
            func = getattr(self, '_emit_' + IR_node.pattern)
            line = func(IR_node)
            return line
        
        input_vars = list()
        for idx, in_edge in enumerate(IR_node.in_edges):
            in_node = self.IR_graph.get_node(in_edge)
            if in_node.type == 'Scope' and len(in_node.return_variables) > 1 and ':' not in in_edge: # the input is a list
                var_name = ', '.join([(in_node.variable_name + "[%s]") %s for s in range(len(in_node.return_variables))])
                input_vars.append(var_name)
            else:
                input_vars.append(self.parent_variable_name(IR_node, [idx]))

        code = "{:<15} = my_{}()([{}])".format(
            IR_node.real_variable_name,
            IR_node.pattern,
            ', '.join(input_vars))
        self._gen_scope_code(IR_node)

        return code


    def _gen_scope_code(self, scope_node):

        def _scope_func(scope_name, params, code, return_var):
            if len(return_var) > 1:
                return_var_code = '[{}]'.format(', '.join(return_var))
                output_shape_code = '        self.output_shapes = [{}]\n'.format(', '.join(['K.int_shape(%s)' %s for s in return_var]))
            else:
                return_var_code = ', '.join(return_var)
                output_shape_code = '        self.output_shapes = K.int_shape({})\n'.format(return_var[0])

            code = """
class my_{}(keras.layers.Layer):
    def __init__(self, **kwargs):
        super(my_{}, self).__init__(**kwargs)
    def call(self, inputs):

{}
{}
{}
        return {}
    
    def compute_output_shape(self, input_shape):
        return self.output_shapes

    """.format(scope_name, scope_name, params, code, output_shape_code, return_var_code)
            return code

        if not self.layers_codes.get(scope_node.pattern, None):
            body_code = str()
            for node_name in scope_node.topology_list:
                node = self.IR_graph.get_node(node_name)
                node_type = node.type

                if hasattr(self, "emit_" + node_type):
                    func = getattr(self, "emit_" + node_type)
                    line = func(node, True)
                    if line != None:
                        body_code += "        " + line + '\n'
                else:
                    print("KerasEmitter has not supported operator [%s]." % (node_type))
                    self.emit_UNKNOWN(node)

            # param_code does not need parameter slice.
            input_params = scope_node.input_params
            
            param_code = str()
            import re
            for i, p in enumerate(scope_node.in_edges):
                p_node = self.IR_graph.get_node(p)
                if p_node.type == 'Scope' and len(p_node.return_variables) > 1 and ':' not in p: # input is a list.
                    param_code += "        {} = [{}]\n".format(p_node.variable_name, ', '.join('inputs[%s]'%s for s in range(i, i + len(p_node.return_variables))))
                else:
                    param_code += "        {} = inputs[{}]\n".format(p_node.variable_name, i) 

            function_code = _scope_func(scope_node.pattern, param_code, body_code, scope_node.return_variables)
            self.layers_codes[scope_node.pattern] = function_code
            return body_code


    def _emit_h_zero(self, IR_node):
        if not self.layers_codes.get(IR_node.pattern, None):
            class_code = '''
class my_h_zero(keras.layers.Layer):
    def __init__(self, **kwargs):
        super(my_h_zero, self).__init__(**kwargs)
    
    def call(self, dummy):
        {:<15} = K.constant(np.full((1, {}), {}))

        return {}
            '''.format(IR_node.variable_name,
            IR_node.get_attr('fill_size'),
            IR_node.get_attr('fill_value'),
            IR_node.variable_name)
            self.layers_codes[IR_node.pattern] = class_code

        code = "{:<15} = my_h_zero()({})".format(IR_node.variable_name, self.parent_variable_name(IR_node))

        return code


    def _layer_Yolo(self):
        self.add_body(0, '''
def yolo_parameter():
    return {}
'''.format(self.yolo_parameter))


    def _layer_Region(self):
        self.add_body(0, '''
def region_parameter():
    return {}
'''.format(self.region_parameter))


    def _layer_SpaceToDepth(self):
        self.add_body(0, '''
def space_to_depth(input, blocksize):
    import tensorflow as tf
    return tf.space_to_depth(input, block_size=blocksize)
''')


    def _layer_Flatten(self):
        self.add_body(0, '''
def __flatten(name, input):
    if input.shape.ndims > 2: return layers.Flatten(name = name)(input)
    else: return input
''')

    def _layer_LRN(self):
        self.add_body(0, '''
from keras.layers.core import Layer
class LRN(Layer):

    def __init__(self, size=5, alpha=0.0005, beta=0.75, k=2, **kwargs):
        self.n = size
        self.alpha = alpha
        self.beta = beta
        self.k = k
        super(LRN, self).__init__(**kwargs)

    def build(self, input_shape):
        self.shape = input_shape
        super(LRN, self).build(input_shape)

    def call(self, x, mask=None):
        half_n = self.n - 1
        squared = K.square(x)
        scale = self.k
        norm_alpha = self.alpha / (2 * half_n + 1)
        if K.image_dim_ordering() == "th":
            b, f, r, c = self.shape
            squared = K.expand_dims(squared, 0)
            squared = K.spatial_3d_padding(squared, padding=((half_n, half_n), (0, 0), (0,0)))
            squared = K.squeeze(squared, 0)
            for i in range(half_n*2+1):
                scale += norm_alpha * squared[:, i:i+f, :, :]
        else:
            b, r, c, f = self.shape
            squared = K.expand_dims(squared, -1)
            squared = K.spatial_3d_padding(squared, padding=((0, 0), (0,0), (half_n, half_n)))
            squared = K.squeeze(squared, -1)
            for i in range(half_n*2+1):
                scale += norm_alpha * squared[:, :, :, i:i+f]

        scale = K.pow(scale, self.beta)
        return x / scale

    def compute_output_shape(self, input_shape):
        return input_shape''')


    def _layer_Conv(self):
        self.add_body(0, """
def convolution(weights_dict, name, input, group, conv_type, filters=None, **kwargs):
    if not conv_type.startswith('layer'):
        layer = keras.applications.mobilenet.DepthwiseConv2D(name=name, **kwargs)(input)
        return layer
    elif conv_type == 'layers.DepthwiseConv2D':
        layer = layers.DepthwiseConv2D(name=name, **kwargs)(input)
        return layer
    
    inp_filters = K.int_shape(input)[-1]
    inp_grouped_channels = int(inp_filters / group)
    out_grouped_channels = int(filters / group)
    group_list = []
    if group == 1:
        func = getattr(layers, conv_type.split('.')[-1])
        layer = func(name = name, filters = filters, **kwargs)(input)
        return layer
    weight_groups = list()
    if not weights_dict == None:
        w = np.array(weights_dict[name]['weights'])
        weight_groups = np.split(w, indices_or_sections=group, axis=-1)
    for c in range(group):
        x = layers.Lambda(lambda z: z[..., c * inp_grouped_channels:(c + 1) * inp_grouped_channels])(input)
        x = layers.Conv2D(name=name + "_" + str(c), filters=out_grouped_channels, **kwargs)(x)
        weights_dict[name + "_" + str(c)] = dict()
        weights_dict[name + "_" + str(c)]['weights'] = weight_groups[c]
        group_list.append(x)
    layer = layers.concatenate(group_list, axis = -1)
    if 'bias' in weights_dict[name]:
        b = K.variable(weights_dict[name]['bias'], name = name + "_bias")
        layer = layer + b
    return layer""")

    def _layer_Scale(self):
        self.add_body(0, """
from keras.engine import Layer, InputSpec
from keras import initializers
from keras  import backend as K


class Scale(Layer):

    def __init__(self,
                 axis=-1,
                 center=True,
                 scale=True,
                 beta_initializer='zeros',
                 gamma_initializer='ones',
                 **kwargs):
        super(Scale, self).__init__(**kwargs)
        self.supports_masking = True
        self.axis = axis
        self.center = center
        self.scale = scale
        self.beta_initializer = initializers.get(beta_initializer)
        self.gamma_initializer = initializers.get(gamma_initializer)


    def build(self, input_shape):
        dim = input_shape[self.axis]
        if dim is None:
            raise ValueError('Axis ' + str(self.axis) + ' of '
                             'input tensor should have a defined dimension '
                             'but the layer received an input with shape ' +
                             str(input_shape) + '.')
        self.input_spec = InputSpec(ndim=len(input_shape),
                                    axes={self.axis: dim})
        shape = (dim,)

        if self.scale:
            self.gamma = self.add_weight(shape=shape,
                                         name='gamma',
                                         initializer=self.gamma_initializer)
        else:
            self.gamma = None
        if self.center:
            self.beta = self.add_weight(shape=shape,
                                        name='beta',
                                        initializer=self.beta_initializer)
        else:
            self.beta = None


        self.built = True

    def call(self, inputs, training=None):
        input_shape = K.int_shape(inputs)
        # Prepare broadcasting shape.
        ndim = len(input_shape)
        reduction_axes = list(range(len(input_shape)))
        del reduction_axes[self.axis]
        broadcast_shape = [1] * len(input_shape)
        broadcast_shape[self.axis] = input_shape[self.axis]
        return K.reshape(self.gamma, broadcast_shape) * inputs + K.reshape(self.beta, broadcast_shape)

    def get_config(self):
        config = {
            'axis': self.axis,
            'center': self.center,
            'scale': self.scale,
        }
        base_config = super(Scale, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape""")


    def _layer_Split(self):
        self.add_body(0, '''
def __split(input, split_num, axis):
    return Lambda(lambda x: tf.split(x, split_num, axis))(input)
        ''')


    def _layer_Unsqueeze(self):
        self.add_body(0, '''
def __unsqueeze(input, axis):
    return Lambda(lambda x: tf.expand_dims(x, axis))(input)
        ''')


    def _layer_Fill(self):
        self.add_body(0, '''
def __fill(input, value):
    class Fill(keras.layers.Layer):
        def call(self, input):
            if keras.backend.backend() =='tensorflow':
                output = tf.fill(input, value)
            else:
                raise NotImplementedError
            self.output_dim = [dim.value for dim in output.shape]
            return output
        
        def compute_output_shape(self, input_shape):
            return tuple(self.output_dim)
    # output = Lambda(lambda x: tf.fill(x, value))(input)
    output = Fill()(input)
    # return output

        ''')


    def _layer_Slice(self):
        self.add_body(0, '''
def __slice(input, start, end, **kargs):
    return Lambda(lambda x: tf.strided_slice(x, start, end, **kargs))(input)
        ''')


    def _layer_Unstack(self):
        self.add_body(0, '''
def __unstack(input, num, axis):
    return Lambda(lambda x: tf.unstack(x, num, axis))(input)
        ''')


    def _layer_Mul(self):
        self.add_body(0, '''
class my_mul(keras.layers.Layer):
    def __init__(self, **kwargs):
        super(my_mul, self).__init__(**kwargs)
    def call(self, inputs):
        res = inputs[0] * inputs[1]
        self.output_shapes = K.int_shape(res)
        return res
    
    def compute_output_shape(self, input_shape):
        return self.output_shapes
''')


    def _layer_Add(self):
        self.add_body(0, '''
class my_add(keras.layers.Layer):
    def __init__(self, **kwargs):
        super(my_add, self).__init__(**kwargs)
    def call(self, inputs):
        res = inputs[0] + inputs[1]
        self.output_shapes = K.int_shape(res)
        return res
    
    def compute_output_shape(self, input_shape):
        return self.output_shapes
''')

    def _layer_Sub(self):
        self.add_body(0, '''
class my_sub(keras.layers.Layer):
    def __init__(self, **kwargs):
        super(my_sub, self).__init__(**kwargs)
    def call(self, inputs):
        res = inputs[0] - inputs[1]
        self.output_shapes = K.int_shape(res)
        return res
    
    def compute_output_shape(self, input_shape):
        return self.output_shapes
''')

    def _layer_Shape(self):
        self.add_body(0, '''
def __shape(input):
    return Lambda(lambda x: tf.shape(x))(input)
        ''')

#     def _layer_Constant(self):
#         self.add_body(0, '''
# class my_constant(keras.layers.Layer):
#     def __init__(self, value, **kwargs):
#         super(my_constant, self).__init__(**kwargs)
#         self._value = value
#     # the input is dummy, just for creating keras graph.
#     def call(self, dummy):
#         res = K.constant(self._value)
#         self.output_shapes = K.int_shape(res)
#         return res
    
#     def compute_output_shape(self, input_shape):
#         return self.output_shapes
# ''')

    def _layer_Mul_Constant(self):
        self.add_body(0, '''
def mul_constant(weight_factor, layer_name):
    weight = Lambda(lambda x: x*weight_factor)
    weight(layer_name)
    return weight.output
''')


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from collections import namedtuple
import numpy as np
from mmdnn.conversion.examples.imagenet_test import TestKit
import mxnet as mx

Batch = namedtuple('Batch', ['data'])


class TestMXNet(TestKit):

    def __init__(self):
        super(TestMXNet, self).__init__()

        self.truth['tensorflow']['inception_v3'] = [(22, 9.6691055), (24, 4.3524752), (25, 3.5957956), (132, 3.5657482), (23, 3.3462858)]
        self.truth['keras']['inception_v3'] = [(21, 0.93430501), (23, 0.0028834261), (131, 0.0014781745), (24, 0.0014518937), (22, 0.0014435325)]

        self.model = self.MainModel.RefactorModel()
        self.model = self.MainModel.deploy_weight(self.model, self.args.w)


    def preprocess(self, image_path):
        self.data = super(TestMXNet, self).preprocess(image_path)
        self.data = np.swapaxes(self.data, 0, 2)
        self.data = np.swapaxes(self.data, 1, 2)
        self.data = np.expand_dims(self.data, 0)


    def print_result(self):
        self.model.forward(Batch([mx.nd.array(self.data)]))
        prob = self.model.get_outputs()[0].asnumpy()
        super(TestMXNet, self).print_result(prob)


    def inference(self, image_path):
        self.preprocess(image_path)

        # self.print_intermediate_result('pooling0', False)

        self.print_result()

        self.test_truth()


    def print_intermediate_result(self, layer_name, if_transpose = False):
        internals = self.model.symbol.get_internals()
        intermediate_output = internals[layer_name + "_output"]
        test_model = mx.mod.Module(symbol=intermediate_output, context=mx.cpu(), data_names=['data'])
        if self.args.preprocess == 'vgg19' or self.args.preprocess == 'inception_v1':
            test_model.bind(for_training=False, data_shapes = [('data', (1, 3, 224, 224))])
        elif 'resnet' in self.args.preprocess or self.args.preprocess == 'inception_v3':
            test_model.bind(for_training=False, data_shapes = [('data', (1, 3, 299, 299))])
        else:
            assert False

        arg_params, aux_params = self.model.get_params()

        test_model.set_params(arg_params = arg_params, aux_params = aux_params, allow_missing = True, allow_extra = True)
        test_model.forward(Batch([mx.nd.array(self.data)]))
        intermediate_output = test_model.get_outputs()[0].asnumpy()

        super(TestMXNet, self).print_intermediate_result(intermediate_output, if_transpose)


    def dump(self, path = None):
        if path is None: path = self.args.dump
        self.model.save_checkpoint(path, 0)
        print ('MXNet checkpoint file is saved with prefix [{}] and iteration 0, generated by [{}.py] and [{}].'.format(
            path, self.args.n, self.args.w))


if __name__ == '__main__':
    tester = TestMXNet()
    if tester.args.dump:
        tester.dump()
    else:
        tester.inference(tester.args.image)



# ----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
# ----------------------------------------------------------------------------------------------

from mmdnn.conversion.common.DataStructure.parser import Parser
from mmdnn.conversion.onnx.onnx_graph import ONNXGraph


class ONNXParser(Parser):
    skip_type = set()

    @property
    def src_graph(self):
        return self.onnx_graph

    @staticmethod
    def _load_model(model_file):
        """Load a ONNX model file from disk

        Parameters
        ----------
        model_file: str
            Path where the model file path is (protobuf file)

        Returns
        -------
        model: A ONNX protobuf model
        """
        from onnx import onnx_pb2
        from mmdnn.conversion.common.IR.IR_graph import load_protobuf_from_file

        model = onnx_pb2.ModelProto()
        load_protobuf_from_file(model, model_file)

        print("ONNX model file [%s] loaded successfully." % model_file)
        return model

    def __init__(self, model_file):
        super(ONNXParser, self).__init__()

        model = ONNXParser._load_model(model_file)
        self.onnx_graph = ONNXGraph(model)
        self.onnx_graph.build()
        self.weight_loaded = True

    def rename_UNKNOWN(self, source_node):
        if source_node.type in self.skip_type:
            return
        print("ONNX has not supported operator [%s] with name [%s]."
              % (source_node.type, source_node.name))
        return

    def gen_IR(self):
        # if node len(in_edges), generate additional DataInput node

        # print
        for layer in self.src_graph.topological_sort:
            current_node = self.src_graph.get_node(layer)
            node_type = current_node.type
            if hasattr(self, "rename_" + node_type):
                func = getattr(self, "rename_" + node_type)
                func(current_node)
            else:
                self.rename_UNKNOWN(current_node)



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
from six import text_type as _text_type
import mxnet as mx
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.common.utils import download_file
from collections import namedtuple
Batch = namedtuple('Batch', ['data'])

network_name_key = ['resnet', 'vgg19', 'squeezenet', 'inception-bn', 'resnext']

_base_model_url = 'http://data.mxnet.io/models/'
_default_model_info = {
    'imagenet1k-inception-bn'           : {'symbol'     : _base_model_url+'imagenet/inception-bn/Inception-BN-symbol.json',
                                           'params'     : _base_model_url+'imagenet/inception-bn/Inception-BN-0126.params',
                                           'image_size' : 224},
    'imagenet1k-resnet-18'              : {'symbol'     : _base_model_url+'imagenet/resnet/18-layers/resnet-18-symbol.json',
                                           'params'     : _base_model_url+'imagenet/resnet/18-layers/resnet-18-0000.params',
                                           'image_size' : 224},
    'imagenet1k-resnet-34'              : {'symbol'     : _base_model_url+'imagenet/resnet/34-layers/resnet-34-symbol.json',
                                           'params'     : _base_model_url+'imagenet/resnet/34-layers/resnet-34-0000.params',
                                           'image_size' : 224},
    'imagenet1k-resnet-50'              : {'symbol'     : _base_model_url+'imagenet/resnet/50-layers/resnet-50-symbol.json',
                                           'params'     : _base_model_url+'imagenet/resnet/50-layers/resnet-50-0000.params',
                                           'image_size' : 224},
    'imagenet1k-resnet-101'             : {'symbol'     : _base_model_url+'imagenet/resnet/101-layers/resnet-101-symbol.json',
                                           'params'     : _base_model_url+'imagenet/resnet/101-layers/resnet-101-0000.params',
                                           'image_size' : 224},
    'imagenet1k-resnet-152'             : {'symbol'     : _base_model_url+'imagenet/resnet/152-layers/resnet-152-symbol.json',
                                           'params'     : _base_model_url+'imagenet/resnet/152-layers/resnet-152-0000.params',
                                           'image_size' : 224},
    'imagenet1k-resnext-50'             : {'symbol'     : _base_model_url+'imagenet/resnext/50-layers/resnext-50-symbol.json',
                                           'params'     : _base_model_url+'imagenet/resnext/50-layers/resnext-50-0000.params',
                                           'image_size' : 224},
    'imagenet1k-resnext-101'            : {'symbol'     : _base_model_url+'imagenet/resnext/101-layers/resnext-101-symbol.json',
                                           'params'     : _base_model_url+'imagenet/resnext/101-layers/resnext-101-0000.params',
                                           'image_size' : 224},
    'imagenet1k-resnext-101-64x4d'      : {'symbol'     : _base_model_url+'imagenet/resnext/101-layers/resnext-101-64x4d-symbol.json',
                                           'params'     : _base_model_url+'imagenet/resnext/101-layers/resnext-101-64x4d-0000.params',
                                           'image_size' : 224},
    'imagenet11k-resnet-152'            : {'symbol'     : _base_model_url+'imagenet-11k/resnet-152/resnet-152-symbol.json',
                                           'params'     : _base_model_url+'imagenet-11k/resnet-152/resnet-152-0000.params',
                                           'image_size' : 224},
    'imagenet11k-place365ch-resnet-152' : {'symbol'     : _base_model_url+'imagenet-11k-place365-ch/resnet-152-symbol.json',
                                           'params'     : _base_model_url+'imagenet-11k-place365-ch/resnet-152-0000.params',
                                           'image_size' : 224},
    'imagenet11k-place365ch-resnet-50'  : {'symbol'     : _base_model_url+'imagenet-11k-place365-ch/resnet-50-symbol.json',
                                           'params'     : _base_model_url+'imagenet-11k-place365-ch/resnet-50-0000.params',
                                           'image_size' : 224},
    'vgg19'                             : {'symbol'     : _base_model_url+'imagenet/vgg/vgg19-symbol.json',
                                           'params'     : _base_model_url+'imagenet/vgg/vgg19-0000.params',
                                           'image_size' : 224},
    'vgg16'                             : {'symbol'     : _base_model_url+'imagenet/vgg/vgg16-symbol.json',
                                           'params'     : _base_model_url+'imagenet/vgg/vgg16-0000.params',
                                           'image_size' : 224},
    'squeezenet_v1.0'                   : {'symbol'     : _base_model_url+'imagenet/squeezenet/squeezenet_v1.0-symbol.json',
                                           'params'     : _base_model_url+'imagenet/squeezenet/squeezenet_v1.0-0000.params',
                                           'image_size' : 224},
    'squeezenet_v1.1'                   : {'symbol'     : _base_model_url+'imagenet/squeezenet/squeezenet_v1.1-symbol.json',
                                           'params'     : _base_model_url+'imagenet/squeezenet/squeezenet_v1.1-0000.params',
                                           'image_size' : 224}
}


def _search_preprocess_key(original_network_name):
    import re
    for key in network_name_key:
        if re.search(key, original_network_name):
            return key
    raise ValueError('preprocess module cannot support [{}]'.format(original_network_name))


def _main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-n', '--network', type=_text_type, help='Model Type', required=True,
                        choices=_default_model_info.keys())

    parser.add_argument('-i', '--image', default=None,
                        type=_text_type, help='Test Image Path')

    parser.add_argument('-o', '--output_dir', default='./',
                        type=_text_type, help='Tensorflow Checkpoint file name')

    args = parser.parse_args()

    if not download_file(_default_model_info[args.network]['symbol'], directory=args.output_dir):
        return -1

    if not download_file(_default_model_info[args.network]['params'], directory=args.output_dir):
        return -1

    print("Model {} saved.".format(args.network))

    file_name = _default_model_info[args.network]['params'].split('/')[-1]
    prefix, epoch_num = file_name[:-7].rsplit('-', 1)

    sym, arg_params, aux_params = mx.model.load_checkpoint(args.output_dir + prefix, int(epoch_num))
    model = mx.mod.Module(symbol=sym)
    model.bind(for_training=False,
               data_shapes=[('data', (1, 3, _default_model_info[args.network]['image_size'],
                                      _default_model_info[args.network]['image_size']))])
    model.set_params(arg_params, aux_params, allow_missing=True, allow_extra=True)

    if args.image:
        import numpy as np

        # need to be updated
        network = _search_preprocess_key(args.network)

        func = TestKit.preprocess_func['mxnet'][network]
        img = func(args.image)
        img = np.swapaxes(img, 0, 2)
        img = np.swapaxes(img, 1, 2)
        img = np.expand_dims(img, axis=0)

        model.forward(Batch([mx.nd.array(img)]))
        predict = model.get_outputs()[0].asnumpy()
        predict = np.squeeze(predict)
        top_indices = predict.argsort()[-5:][::-1]
        result = [(i, predict[i]) for i in top_indices]
        print(result)

    return 0


if __name__ == '__main__':
    _main()



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
import os
from six import text_type as _text_type
from mmdnn.conversion.common.utils import download_file
import paddle.v2 as paddle
import gzip
from paddle.trainer_config_helpers.config_parser_utils import \
    reset_parser

BASE_MODEL_URL = 'http://cloud.dlnel.org/filepub/?uuid='
# pylint: disable=line-too-long
MODEL_URL = {
    'resnet50'             : BASE_MODEL_URL + 'f63f237a-698e-4a22-9782-baf5bb183019',
    'resnet101'            : BASE_MODEL_URL + '3d5fb996-83d0-4745-8adc-13ee960fc55c',
    'vgg16'                : BASE_MODEL_URL + 'aa0e397e-474a-4cc1-bd8f-65a214039c2e',
}
# pylint: enable=line-too-long
IMG_SIZE = 224
CLASS_DIMS = {
        'resnet50'             : 1000,
        'resnet101'            : 1000,
        'vgg16'                : 1001, # work at 1001, but fail at 1000
        'alexnet'              : 1001,
}

def dump_v2_config(topology, save_path, binary=False):
    import collections

    from paddle.trainer_config_helpers.layers import LayerOutput
    from paddle.v2.layer import parse_network
    from paddle.proto import TrainerConfig_pb2
    """ Dump the network topology to a specified file.
    This function is only used to dump network defined by using PaddlePaddle V2
    API.
    :param topology: The output layers in the entire network.
    :type topology: LayerOutput|List|Tuple
    :param save_path: The path to save the dump network topology.
    :type save_path: str
    :param binary: Whether to dump the serialized network topology. The default
                value is false.
    :type binary: bool.
    """

    if isinstance(topology, LayerOutput):
        topology = [topology]
    elif isinstance(topology, collections.Sequence):
        for out_layer in topology:
            assert isinstance(out_layer, LayerOutput), (
                "The type of each element in the parameter topology "
                "should be LayerOutput.")
    else:
        raise RuntimeError("Error input type for parameter topology.")

    model_str = parse_network(topology)
    with open(save_path, "w") as fout:
        if binary:
            fout.write(model_str.SerializeToString())
        else:
            fout.write(str(model_str))

def _main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-n', '--network', type=_text_type, help='Model Type', required=True,
                        choices=MODEL_URL.keys())

    parser.add_argument('-i', '--image', default=None,
                        type=_text_type, help='Test Image Path')

    parser.add_argument('-o', '--output_dir', default='./',
                        type=_text_type, help='Paddlepaddle parameters file name')

    args = parser.parse_args()

    fn = download_file(MODEL_URL[args.network], local_fname = architecture + '.tar.gz', directory=args.output_dir)
    if not fn:
        return -1


    DATA_DIM = 3 * IMG_SIZE * IMG_SIZE  # Use 3 * 331 * 331 or 3 * 299 * 299 for Inception-ResNet-v2.
    CLASS_DIM = CLASS_DIMS[args.network]

    # refer to https://github.com/PaddlePaddle/Paddle/blob/develop/python/paddle/v2/tests/test_rnn_layer.py#L35
    reset_parser()

    # refer to https://github.com/PaddlePaddle/Paddle/issues/7403
    paddle.init(use_gpu=False, trainer_count=1)

    image = paddle.layer.data(
        name="image", type=paddle.data_type.dense_vector(DATA_DIM))
    if 'resnet' in architecture:
        from mmdnn.conversion.examples.paddle.models import resnet
        depth = int(architecture.strip('resnet'))
        out = resnet.resnet_imagenet(image, class_dim=CLASS_DIM, depth=depth)
    elif architecture == 'vgg16':
        from mmdnn.conversion.examples.paddle.models import vgg
        out = vgg.vgg16(image, class_dim=CLASS_DIM)
    else:
        print("Not support for {} yet.", architecture)
        return None

    dump_v2_config(out, args.output_dir + architecture + '.bin')


    print("Model {} is saved as {} and {}.".format(args.network,  args.output_dir + architecture + '.bin', fn))

    if args.image:

        import numpy as np
        from mmdnn.conversion.examples.imagenet_test import TestKit
        func = TestKit.preprocess_func['paddle'][args.network]
        img = func(args.image)
        img = np.transpose(img, (2, 0, 1))
        test_data = [(img.flatten(),)]

        with gzip.open(parameters_file, 'r') as f:
            parameters = paddle.parameters.Parameters.from_tar(f)

        predict = paddle.infer(output_layer = out, parameters=parameters, input=test_data)
        predict = np.squeeze(predict)
        top_indices = predict.argsort()[-5:][::-1]
        result = [(i, predict[i]) for i in top_indices]
        print(result)
        print(np.sum(result))

    return 0


if __name__ == '__main__':
    _main()



import torch
from collections import OrderedDict

def parse_cfg(cfgfile):
    def erase_comment(line):
        line = line.split('#')[0]
        return line
    blocks = []
    fp = open(cfgfile, 'r')
    block =  None
    line = fp.readline()
    while line != '':
        line = line.rstrip()
        if line == '' or line[0] == '#':
            line = fp.readline()
            continue        
        elif line[0] == '[':
            if block:
                blocks.append(block)
            block = OrderedDict()
            block['type'] = line.lstrip('[').rstrip(']')
            # set default value
            if block['type'] == 'convolutional':
                block['batch_normalize'] = 0
        else:
            line = erase_comment(line)
            key,value = line.split('=')
            key = key.strip()
            if key == 'type':
                key = '_type'
            value = value.strip()
            block[key] = value
        line = fp.readline()

    if block:
        blocks.append(block)
    fp.close()
    return blocks

def print_cfg(blocks):
    for block in blocks:
        print('[%s]' % (block['type']))
        for key,value in block.items():
            if key != 'type':
                print('%s=%s' % (key, value))
        print('')
def save_cfg(blocks, cfgfile):
    with open(cfgfile, 'w') as fp:
        for block in blocks:
            fp.write('[%s]\n' % (block['type']))
            for key,value in block.items():
                if key != 'type':
                    fp.write('%s=%s\n' % (key, value))
            fp.write('\n')

def print_cfg_nicely(blocks):
    print('layer     filters    size              input                output');
    prev_width = 416
    prev_height = 416
    prev_filters = 3
    out_filters =[]
    out_widths =[]
    out_heights =[]
    ind = -2
    for block in blocks:
        ind = ind + 1
        if block['type'] == 'net':
            prev_width = int(block['width'])
            prev_height = int(block['height'])
            continue
        elif block['type'] == 'convolutional':
            filters = int(block['filters'])
            kernel_size = int(block['size'])
            stride = int(block['stride'])
            is_pad = int(block['pad'])
            pad = (kernel_size-1)/2 if is_pad else 0
            width = (prev_width + 2*pad - kernel_size)/stride + 1
            height = (prev_height + 2*pad - kernel_size)/stride + 1
            print('%5d %-6s %4d  %d x %d / %d   %3d x %3d x%4d   ->   %3d x %3d x%4d' % (ind, 'conv', filters, kernel_size, kernel_size, stride, prev_width, prev_height, prev_filters, width, height, filters))
            prev_width = width
            prev_height = height
            prev_filters = filters
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'maxpool':
            pool_size = int(block['size'])
            stride = int(block['stride'])
            width = prev_width/stride
            height = prev_height/stride
            print('%5d %-6s       %d x %d / %d   %3d x %3d x%4d   ->   %3d x %3d x%4d' % (ind, 'max', pool_size, pool_size, stride, prev_width, prev_height, prev_filters, width, height, filters))
            prev_width = width
            prev_height = height
            prev_filters = filters
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'avgpool':
            width = 1
            height = 1
            print('%5d %-6s                   %3d x %3d x%4d   ->      %3d' % (ind, 'avg', prev_width, prev_height, prev_filters,  prev_filters))
            prev_width = 1
            prev_height = 1
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'softmax':
            print('%5d %-6s                                    ->      %3d' % (ind, 'softmax', prev_filters))
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'cost':
            print('%5d %-6s                                     ->      %3d' % (ind, 'cost', prev_filters))
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'reorg':
            stride = int(block['stride'])
            filters = stride * stride * prev_filters
            width = prev_width/stride
            height = prev_height/stride
            print('%5d %-6s             / %d   %3d x %3d x%4d   ->   %3d x %3d x%4d' % (ind, 'reorg', stride, prev_width, prev_height, prev_filters, width, height, filters))
            prev_width = width
            prev_height = height
            prev_filters = filters
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'route':
            layers = block['layers'].split(',')
            layers = [int(i) if int(i) > 0 else int(i)+ind for i in layers]
            if len(layers) == 1:
                print('%5d %-6s %d' % (ind, 'route', layers[0]))
                prev_width = out_widths[layers[0]]
                prev_height = out_heights[layers[0]]
                prev_filters = out_filters[layers[0]]
            elif len(layers) == 2:
                print('%5d %-6s %d %d' % (ind, 'route', layers[0], layers[1]))
                prev_width = out_widths[layers[0]]
                prev_height = out_heights[layers[0]]
                assert(prev_width == out_widths[layers[1]])
                assert(prev_height == out_heights[layers[1]])
                prev_filters = out_filters[layers[0]] + out_filters[layers[1]]
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'region':
            print('%5d %-6s' % (ind, 'detection'))
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'shortcut':
            from_id = int(block['from'])
            from_id = from_id if from_id > 0 else from_id+ind
            print('%5d %-6s %d' % (ind, 'shortcut', from_id))
            prev_width = out_widths[from_id]
            prev_height = out_heights[from_id]
            prev_filters = out_filters[from_id]
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'softmax':
            print('%5d %-6s' % (ind, 'softmax'))
            out_widths.append(prev_width)
            out_heights.append(prev_height)
            out_filters.append(prev_filters)
        elif block['type'] == 'connected':
            filters = int(block['output'])
            print('%5d %-6s                            %d  ->      %3d' % (ind, 'connected', prev_filters,  filters))
            prev_filters = filters
            out_widths.append(1)
            out_heights.append(1)
            out_filters.append(prev_filters)
        else:
            print('unknown type %s' % (block['type']))

def load_conv(buf, start, conv_model):
    num_w = conv_model.weight.numel()
    num_b = conv_model.bias.numel()
    conv_model.bias.data.copy_(torch.from_numpy(buf[start:start+num_b]));   start = start + num_b
    conv_model.weight.data.copy_(torch.from_numpy(buf[start:start+num_w])); start = start + num_w
    return start

def save_conv(fp, conv_model):
    if conv_model.bias.is_cuda:
        convert2cpu(conv_model.bias.data).numpy().tofile(fp)
        convert2cpu(conv_model.weight.data).numpy().tofile(fp)
    else:
        conv_model.bias.data.numpy().tofile(fp)
        conv_model.weight.data.numpy().tofile(fp)

def load_conv_bn(buf, start, conv_model, bn_model):
    num_w = conv_model.weight.numel()
    num_b = bn_model.bias.numel()
    bn_model.bias.data.copy_(torch.from_numpy(buf[start:start+num_b]));     start = start + num_b
    bn_model.weight.data.copy_(torch.from_numpy(buf[start:start+num_b]));   start = start + num_b
    bn_model.running_mean.copy_(torch.from_numpy(buf[start:start+num_b]));  start = start + num_b
    bn_model.running_var.copy_(torch.from_numpy(buf[start:start+num_b]));   start = start + num_b
    conv_model.weight.data.copy_(torch.from_numpy(buf[start:start+num_w])); start = start + num_w 
    return start

def save_conv_bn(fp, conv_model, bn_model):
    if bn_model.bias.is_cuda:
        convert2cpu(bn_model.bias.data).numpy().tofile(fp)
        convert2cpu(bn_model.weight.data).numpy().tofile(fp)
        convert2cpu(bn_model.running_mean).numpy().tofile(fp)
        convert2cpu(bn_model.running_var).numpy().tofile(fp)
        convert2cpu(conv_model.weight.data).numpy().tofile(fp)
    else:
        bn_model.bias.data.numpy().tofile(fp)
        bn_model.weight.data.numpy().tofile(fp)
        bn_model.running_mean.numpy().tofile(fp)
        bn_model.running_var.numpy().tofile(fp)
        conv_model.weight.data.numpy().tofile(fp)

def save_conv_shrink_bn(fp, conv_model, bn_model, eps=1e-5):
    if bn_model.bias.is_cuda:
        bias = bn_model.bias.data - bn_model.running_mean * bn_model.weight.data / torch.sqrt(bn_model.running_var + eps)
        convert2cpu(bias).numpy().tofile(fp)
        s = conv_model.weight.data.size()
        weight = conv_model.weight.data * (bn_model.weight.data / torch.sqrt(bn_model.running_var + eps)).view(-1,1,1,1).repeat(1, s[1], s[2], s[3])
        convert2cpu(weight).numpy().tofile(fp)
    else:
        bias = bn_model.bias.data - bn_model.running_mean * bn_model.weight.data / torch.sqrt(bn_model.running_var + eps)
        bias.numpy().tofile(fp)
        s = conv_model.weight.data.size()
        weight = conv_model.weight.data * (bn_model.weight.data / torch.sqrt(bn_model.running_var + eps)).view(-1,1,1,1).repeat(1, s[1], s[2], s[3])
        weight.numpy().tofile(fp)

def load_fc(buf, start, fc_model):
    num_w = fc_model.weight.numel()
    num_b = fc_model.bias.numel()
    fc_model.bias.data.copy_(torch.from_numpy(buf[start:start+num_b]));     start = start + num_b
    fc_model.weight.data.copy_(torch.from_numpy(buf[start:start+num_w]));   start = start + num_w 
    return start

def save_fc(fp, fc_model):
    fc_model.bias.data.numpy().tofile(fp)
    fc_model.weight.data.numpy().tofile(fp)


if __name__ == '__main__':
    import sys
    if len(sys.argv) != 2:
        print('Usage: python cfg.py model.cfg')
        exit()

    blocks = parse_cfg(sys.argv[1])
    print_cfg_nicely(blocks)



import tensorflow as tf
from tensorflow.python.platform import gfile
import os
import shutil
import sys

def visualize(model_filename, log_dir):
    with tf.Session() as sess:
        tf.train.import_meta_graph(model_filename)
        train_writer = tf.summary.FileWriter(log_dir)
        train_writer.add_graph(sess.graph)
        train_writer.close()

def _main():
    """
    Visualize the frozen TF graph using tensorboard.

    Arguments
    ----------
    - path to the checkpoint meta file (.ckpt.meta)
    - path to a log directory for writing graph summary for visualization

    Usage
    ----------
    python vis_meta.py model.ckpt.meta /tmp/pb


    To kill a previous tensorboard process, use the following commands in the terminal
    ps aux | grep tensorboard
    kill PID
    """

    if len(sys.argv) != 3:
        raise ValueError("Usage: python vis_meta.py /path/to/model.meta /path/to/log/directory")
    # load file
    visualize(sys.argv[1], sys.argv[2])
    os.system("tensorboard --logdir=" + sys.argv[2])


if __name__ == "__main__":
    _main()


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import
import os
import coremltools
from coremltools.models import MLModel
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.examples.extractor import base_extractor
from mmdnn.conversion.common.utils import download_file


class coreml_extractor(base_extractor):

    _base_model_url = "https://docs-assets.developer.apple.com/coreml/models/"

    # from collections import namedtuple
    # Batch = namedtuple('Batch', ['data'])

    # TODO
    # Apple has published some of their own models. They can be downloaded from https://developer.apple.com/machine-learning/.
    # Those published models are: SqueezeNet, Places205-GoogLeNet, ResNet50, Inception v3, VGG16
    architecture_map = {
        'inception_v3'      : "https://docs-assets.developer.apple.com/coreml/models/Inceptionv3.mlmodel",
        'vgg16'             : "https://docs-assets.developer.apple.com/coreml/models/VGG16.mlmodel",
        'vgg19'             : None,
        'resnet50'            : "https://docs-assets.developer.apple.com/coreml/models/Resnet50.mlmodel",  # resnet50
        'mobilenet'         : "https://docs-assets.developer.apple.com/coreml/models/MobileNet.mlmodel",
        'xception'          : None,
        'inception_resnet'  : None,
        'densenet'          : None,
        'nasnet'            : None,
        'tinyyolo'          : "https://s3-us-west-2.amazonaws.com/coreml-models/TinyYOLO.mlmodel"

    }



    @classmethod
    def download(cls, architecture, path = './'):
        if cls.sanity_check(architecture):
            architecture_file = download_file(cls.architecture_map[architecture], directory = path)
            if not architecture_file:
                return None


            print('Coreml model {} is saved in [{}]'.format(architecture, path))
            return architecture_file
        else:
            return None


    @classmethod
    def inference(cls, architecture, model_path, image_path):
        # TODO
        from PIL import Image
        import numpy as np
        from coremltools.models._infer_shapes_nn_mlmodel import infer_shapes
        if cls.sanity_check(architecture):
            func = TestKit.preprocess_func['coreml'][architecture]


            import inspect
            funcstr = inspect.getsource(func)

            if len(funcstr.split(',')) == 3:
                size = int(funcstr.split('path,')[1].split(')')[0])
            else:
                size = int(funcstr.split('path,')[1].split(',')[0])



            img = Image.open(image_path)
            img = img.resize((size, size))

            # load model
            model = MLModel(model_path)
            spec = model.get_spec()

            # TODO: Multiple inputs
            input_name = spec.description.input[0].name

            # TODO: Multiple outputs
            output_name = spec.description.output[0].name

            # inference
            input_data = img
            coreml_input = {input_name: img}
            coreml_output = model.predict(coreml_input)


            prob = coreml_output[output_name]
            if isinstance(prob, dict):
                prob = list(coreml_output[output_name].values())
            prob = np.array(prob).squeeze()

            return prob

        else:
            return None






from __future__ import absolute_import
from __future__ import division
import numpy as np

from mmdnn.conversion.caffe.errors import ConversionError
from mmdnn.conversion.caffe.common_graph import Node
from mmdnn.conversion.caffe.network import DEFAULT_PADDING
from mmdnn.conversion.caffe.utils import get_lower_case
from mmdnn.conversion.common.IR.graph_pb2 import TensorShape


def get_handler_name(node_kind):
    if node_kind is None:
        return node_kind
    else:
        if len(node_kind) <= 4:
            return node_kind.lower()
        else:
            return get_lower_case(node_kind)


class NodeMapper(object):

    @classmethod
    def _convert_output_shape(cls, kwargs, node):
        shape = TensorShape()
        dim = shape.dim.add()
        dim.size = -1

        if len(node.output_shape) > 2:
            for i in node.output_shape[2:]:
                dim = shape.dim.add()
                dim.size = i
            dim = shape.dim.add()
            dim.size = node.output_shape.channels
        else:
            dim = shape.dim.add()
            dim.size = node.output_shape[1]
        kwargs['_output_shapes'] = [shape]

    @classmethod
    def get_kernel_params(cls, node, input_shape):
        kwargs = {}

        if node.kernel_parameters.global_pooling:
            kwargs['kernel_shape'] = [1, input_shape.height, input_shape.width, 1]
            kwargs['pads'] = [0] * 8

        else:
            from mmdnn.conversion.caffe.graph import NodeKind
            if node.kind == NodeKind.Pooling:
                kwargs['kernel_shape'] = [1, node.kernel_parameters.k_h, node.kernel_parameters.k_w, 1]
            elif node.kind in [NodeKind.Convolution, NodeKind.Deconvolution]:
                pass
            else:
                raise ValueError

            dilation = node.parameters.dilation[0] if hasattr(node.parameters, 'dilation') and node.parameters.dilation else 1
            o_h_caffe = node.output_shape.height
            o_w_caffe = node.output_shape.width
            ko_h = dilation * (int(node.kernel_parameters.k_h) - 1) + 1
            ko_w = dilation * (int(node.kernel_parameters.k_w) - 1) + 1

            if node.kind == NodeKind.Deconvolution:
                o_h_tf = int(node.kernel_parameters.s_h) * (input_shape.height - 1) + ko_h - 2 * int(node.kernel_parameters.p_h)
                o_w_tf = int(node.kernel_parameters.s_w) * (input_shape.width - 1) + ko_w - 2 * int(node.kernel_parameters.p_w)
            else:
                o_h_tf = (input_shape.height + node.kernel_parameters.p_h * 2 - ko_h + 1) // node.kernel_parameters.s_h
                o_w_tf = (input_shape.width + node.kernel_parameters.p_w * 2 - ko_w + 1) // node.kernel_parameters.s_w

            kwargs['pads'] = [0, node.kernel_parameters.p_h, node.kernel_parameters.p_w, 0] + \
                    [0, node.kernel_parameters.p_h + o_h_caffe - o_h_tf, node.kernel_parameters.p_w + o_w_caffe - o_w_tf, 0]

        kwargs['strides'] = [1, node.kernel_parameters.s_h, node.kernel_parameters.s_w, 1]
        cls._convert_output_shape(kwargs, node)

        return kwargs


    @classmethod
    def map_data(cls, node):
        # TODO: We need to identify whether this is 4D image data, otherwise we shouldn't change the dimension order
        shape = TensorShape()
        dim = shape.dim.add()
        dim.size = -1
        for i in node.output_shape[2:]:
            dim = shape.dim.add()
            dim.size = i
        dim = shape.dim.add()
        dim.size = node.output_shape.channels

        kwargs = {'shape': shape} # Ignore the dimension of batch size
        cls._convert_output_shape(kwargs, node)
        return Node.create('DataInput', **kwargs)


    @classmethod
    def map_input(cls, node):
        return cls.map_data(node)

    @classmethod
    def map_convolution(cls, node):
        parent, _ = node.get_only_parent()
        kwargs = cls.get_kernel_params(node, parent.output_shape)
        kwargs['kernel_shape'] = [node.kernel_parameters.k_h, node.kernel_parameters.k_w, parent.output_shape.channels, node.parameters.num_output]
        kwargs['use_bias'] = node.parameters.bias_term
        if node.parameters.dilation:
            dilation = node.parameters.dilation[0]
            if dilation != 1:
                kwargs['dilations'] = [1, dilation, dilation, 1]
        kwargs['group'] = node.parameters.group
        return Node.create('Conv', **kwargs)


    @classmethod
    def map_deconvolution(cls, node):
        parent, _ = node.get_only_parent()
        kwargs = cls.get_kernel_params(node, parent.output_shape)

        kwargs['kernel_shape'] = [node.kernel_parameters.k_h, node.kernel_parameters.k_w, node.parameters.num_output, parent.output_shape.channels]
        kwargs['use_bias'] = node.parameters.bias_term
        if node.parameters.dilation:
            dilation = node.parameters.dilation[0]
            if dilation != 1:
                kwargs['dilations'] = [1, dilation, dilation, 1]
        kwargs['group'] = node.parameters.group
        return Node.create('ConvTranspose', **kwargs)


    @classmethod
    def map_crop(cls, node):
        kwargs = {}
        cls._convert_output_shape(kwargs, node)
        offset = node.parameters.offset
        if offset:
            if len(offset) == 1:
                kwargs['border'] = [offset[0], offset[0], 0, 0]
            else:
                kwargs['border'] = [offset[0], offset[1], 0, 0]

        return Node.create('Crop', **kwargs)


    @classmethod
    def map_relu(cls, node):
        kwargs = {}
        cls._convert_output_shape(kwargs, node)
        return Node.create('Relu', **kwargs)

    @classmethod
    def map_p_re_lu(cls, node):
        # print(node.parameters)
        # assert False
        try:
            scale_value = float(node.parameters.filler.value)
            kwargs = {'gamma' : scale_value}
        except ConversionError:
            kwargs = {'gamma' : 0.25}
        cls._convert_output_shape(kwargs, node)
        return Node.create('PRelu', **kwargs)


    @classmethod
    def map_pooling(cls, node):
        parent, _ = node.get_only_parent()
        kwargs = cls.get_kernel_params(node, parent.output_shape)
        if node.parameters.pool == 0:
            kwargs['pooling_type'] = 'MAX'
        elif node.parameters.pool == 1:
            kwargs['pooling_type'] = 'AVG'
        else:
            # Stochastic pooling, for instance.
            raise ConversionError('Unsupported pooling type.')
        cls._convert_output_shape(kwargs, node)
        return Node.create('Pool', **kwargs)


    @classmethod
    def _add_flatten_layer(cls, node):
        shape = TensorShape()
        dim = shape.dim.add()
        dim.size = -1

        dim = shape.dim.add()
        dim.size = 1
        for i in node.output_shape[1:]:
            dim.size *= i
        kwargs = {'_output_shapes' : [shape]}
        return Node.create('Flatten', **kwargs)

    @classmethod
    def map_inner_product(cls, node):
        #TODO: Axis
        assert node.parameters.axis == 1
        #TODO: Unbiased
        shape = TensorShape()
        dim = shape.dim.add()
        dim.size = -1
        dim = shape.dim.add()
        dim.size = 1
        for i in node.output_shape[1:]:
            dim.size *= i
        kwargs = {'use_bias' : node.parameters.bias_term, 'units' : node.parameters.num_output,
                '_output_shapes': [shape]}

        # check if need the Flatten layer
        parent, _ = node.get_only_parent()
        ret = []

        # if parent.output_shape.height > 1 or parent.output_shape.width > 1:
        ret.append(cls._add_flatten_layer(parent))
        ret.append(Node.create('FullyConnected', **kwargs))
        return ret

    @classmethod
    def map_softmax(cls, node):
        kwargs = {}
        cls._convert_output_shape(kwargs, node)
        return Node.create('Softmax', **kwargs)

    @classmethod
    def map_lrn(cls, node):
        params = node.parameters
        assert params.local_size % 2 == 1
        kwargs = {'size': int((params.local_size + 1) / 2), 'alpha': params.alpha, 'beta': params.beta, 'k' : params.k}
        cls._convert_output_shape(kwargs, node)
        return Node.create('LRN', **kwargs)

    @classmethod
    def map_concat(cls, node):
        kwargs = {'axis': (2, 3, 1, 0)[node.parameters.axis]}
        cls._convert_output_shape(kwargs, node)
        return Node.create('Concat', **kwargs)

    @classmethod
    def map_dropout(cls, node):
        kwargs = {'keep_prob': node.parameters.dropout_ratio}
        cls._convert_output_shape(kwargs, node)
        return Node.create('Dropout', **kwargs)

    @classmethod
    def map_batch_norm(cls, node):
        kwargs = {'scale' : len(node.data) >= 3, 'bias' : len(node.data) == 4}
        epsilon = node.parameters.eps
        kwargs['epsilon'] = epsilon
        cls._convert_output_shape(kwargs, node)
        return Node.create('BatchNorm', **kwargs)

    @classmethod
    def map_scale(cls, node):
        raise NotImplementedError
        # TODO: The gamma parameter has to be set (in node.data?) and this should work.
        # Also, mean should be set to 0, and var to 1, just to be safe.
        scale_value = float(node.parameters.filler.value)
        kwargs = {'scale' : True, 'bias' : False, 'gamma' : scale_value, 'epsilon': 0}
        return Node.create('BatchNorm', **kwargs)

    @classmethod
    def map_eltwise(cls, node):
        operations = {0: 'Mul', 1: 'Add', 2: 'Max'}
        op_code = node.parameters.operation
        try:
            kwargs = {}
            cls._convert_output_shape(kwargs, node)
            return Node.create(operations[op_code], **kwargs)
        except KeyError:
            raise ConversionError('Unknown elementwise operation: {}'.format(op_code))

    @classmethod
    def map_abs_val(cls, node):
        return Node.create('Abs')

    @classmethod
    def map_tanh(cls, node):
        return Node.create('Tanh')

    @classmethod
    def map_sigmoid(cls, node):
        return Node.create('Sigmoid')

    @classmethod
    def map_reshape(cls, node):
        kwargs = {'shape' : [dim for dim in node.output_shape]}
        cls._convert_output_shape(kwargs, node)
        return Node.create('Reshape', **kwargs)

    @classmethod
    def map_flatten(cls, node):
        return cls._add_flatten_layer(node)

    @classmethod
    def map_split(cls, node):
        # skip the split node
        return

    @classmethod
    def map_elu(cls, node):
        kwargs = {}
        cls._convert_output_shape(kwargs, node)
        return Node.create('ELU', **kwargs)



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import os
import numpy as np
from torch.utils.serialization import load_lua
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.common.DataStructure.parser import Parser
from mmdnn.conversion.torch.torch_graph import TorchGraph


class TorchParser(Parser):

    ############
    # property #
    ############

    @property
    def src_graph(self):
        return self.torch_graph


    ####################
    # Public Functions #
    ####################

    def __init__(self, model_file_name, input_shape):
        super(TorchParser, self).__init__()
        if not os.path.exists(model_file_name):
            raise ValueError("Torch7 model file [{}] is not found.".format(model_file_name))
        model = load_lua(model_file_name)
        if type(model).__name__=='hashable_uniq_dict':
            model = model.model
        model.evaluate()
        self.weight_loaded = True

        # Build network graph
        self.torch_graph = TorchGraph(model)
        self.torch_graph.build([[1] + list(map(int, input_shape))])


    def gen_IR(self):
        print ("OK")
        assert False
        print (self.torch_graph.model.childrens())
        for layer in self.src_graph.topological_sort:
            current_node = self.src_graph.get_node(layer)
            node_type = current_node.type

            if hasattr(self, "rename_" + node_type):
                func = getattr(self, "rename_" + node_type)
                func(current_node)

            else:
                self.rename_UNKNOWN(current_node)

    ##########
    # Layers #
    ##########
    def rename_UNKNOWN(self, source_node):
        print (source_node.layer)
        print (source_node.layer.data.size())
        assert False
        print("PyTorch parser has not supported operator [%s] with name [%s]."
              % (source_node.type, source_node.name))

    def rename_NoneType(self, source_node):
        assert source_node.name in self.src_graph.input_layers
        IR_node = self._convert_identity_operation(source_node, new_op="DataInput")
        for dim in self.input_shape:
            new_dim = IR_node.attr["shape"].shape.dim.add()
            if dim == None:
                new_dim.size = -1
            else:
                new_dim.size = dim


    def rename_ConvNd(self, source_node):
        kwargs = dict()
        kwargs['dilations'] = [1] + list(source_node.get_attr('dilation')) + [1]
        kwargs['pads'] = ([0] + list(source_node.get_attr('padding')) + [0]) * 2
        kwargs['strides'] = [1] + list(source_node.get_attr('stride')) + [1]
        kwargs['group'] = source_node.get_attr('groups')

        # handle weight
        weight = source_node.get_attr('next_functions')[1][0].variable.data.numpy()
        dim = weight.ndim - 2

        if source_node.get_attr('transposed'):
            IR_node = self._convert_identity_operation(source_node, new_op="ConvTranpose")
            weight = np.transpose(weight, list(range(2, dim + 2)) + [0, 1])
        else:
            IR_node = self._convert_identity_operation(source_node, new_op="Conv")
            weight = np.transpose(weight, list(range(2, dim + 2)) + [1, 0])

        self.set_weight(source_node.name, 'weights', weight)
        kwargs['kernel_shape'] = list(weight.shape)

        # handle bias
        if source_node.get_attr('next_functions')[2][0]:
            bias = source_node.get_attr('next_functions')[2][0].variable.data.numpy()
            self.set_weight(source_node.name, 'bias', weight)
            kwargs['use_bias'] = True
        else:
            kwargs['use_bias'] = False

        assign_IRnode_values(IR_node, kwargs)


    def rename_Threshold(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Relu')


    def rename_MaxPool2d(self, source_node):
        self._convert_pooling(source_node)


    def rename_View(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Reshape')
        assign_IRnode_values(IR_node, {'shape' : list(source_node.get_attr('new_sizes'))[1:]})


    def rename_Addmm(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='FullyConnected')
        kwargs = dict()

        # handle weight
        weight = source_node.get_attr('next_functions')[2][0].next_functions[0][0].variable.data.numpy()
        weight = np.transpose(weight)
        kwargs['units'] = weight.shape[1]
        self.set_weight(source_node.name, 'weights', weight)

        # handle bias
        if source_node.get_attr('next_functions')[0][0]:
            bias = source_node.get_attr('next_functions')[0][0].variable.data.numpy()
            kwargs['use_bias'] = True
            self.set_weight(source_node.name, 'bias', weight)

        assign_IRnode_values(IR_node, kwargs)

        print(IR_node)


    ####################
    # Helper Functions #
    ####################

    @staticmethod
    def _copy_and_reop(source_node, IR_node, new_op = None):
        if new_op == None: new_op = source_node.type
        IR_node.name = source_node.name
        IR_node.op = new_op


    def _convert_identity_operation(self, source_node, in_edge_count = None, new_op = None):
        IR_node = self.IR_graph.node.add()
        PyTorchParser._copy_and_reop(source_node, IR_node, new_op)
        self.convert_inedge(source_node, IR_node, 0, in_edge_count)
        return IR_node

    def _convert_pooling(self, source_node):
        kwargs = dict()
        kwargs['strides'] = [1] + list(source_node.get_attr('stride')) + [1]
        kwargs['dilations'] = [1] + list(source_node.get_attr('dilation')) + [1]
        kwargs['pads'] = ([0] + list(source_node.get_attr('padding')) + [0]) * 2
        kwargs['kernel_shape'] = [1] + list(source_node.get_attr('kernel_size')) + [1]
        IR_node = self._convert_identity_operation(source_node, new_op="Pool")

        if source_node.name.startswith('Max'):
            kwargs['pooling_type'] = 'MAX'
        elif source_node.name.startswith('Avg'):
            kwargs['pooling_type'] = 'MAX'
        else:
            raise ValueError('Unknown pooling type')

        assign_IRnode_values(IR_node, kwargs)



import sys as _sys
import google.protobuf.text_format as text_format
from six import text_type as _text_type



def _convert(args):
    if args.dstFramework == 'caffe':
        from mmdnn.conversion.caffe.caffe_emitter import CaffeEmitter
        if args.IRWeightPath is None:
            emitter = CaffeEmitter(args.IRModelPath)
        else:
            assert args.dstWeightPath
            emitter = CaffeEmitter((args.IRModelPath, args.IRWeightPath))

    elif args.dstFramework == 'keras':
        from mmdnn.conversion.keras.keras2_emitter import Keras2Emitter
        emitter = Keras2Emitter((args.IRModelPath, args.IRWeightPath))

    elif args.dstFramework == 'tensorflow':
        from mmdnn.conversion.tensorflow.tensorflow_emitter import TensorflowEmitter
        if args.IRWeightPath is None:
            # Convert network architecture only
            emitter = TensorflowEmitter(args.IRModelPath)
        else:
            emitter = TensorflowEmitter((args.IRModelPath, args.IRWeightPath))

    elif args.dstFramework == 'cntk':
        from mmdnn.conversion.cntk.cntk_emitter import CntkEmitter
        if args.IRWeightPath is None:
            emitter = CntkEmitter(args.IRModelPath)
        else:
            emitter = CntkEmitter((args.IRModelPath, args.IRWeightPath))

    elif args.dstFramework == 'coreml':
        raise NotImplementedError("CoreML emitter is not finished yet.")

    elif args.dstFramework == 'pytorch':
        if not args.dstWeightPath or not args.IRWeightPath:
            raise ValueError("Need to set a target weight filename.")
        from mmdnn.conversion.pytorch.pytorch_emitter import PytorchEmitter
        emitter = PytorchEmitter((args.IRModelPath, args.IRWeightPath))

    elif args.dstFramework == 'mxnet':
        from mmdnn.conversion.mxnet.mxnet_emitter import MXNetEmitter
        if args.IRWeightPath is None:
            emitter = MXNetEmitter(args.IRModelPath)
        else:
            if args.dstWeightPath is None:
                raise ValueError("MXNet emitter needs argument [dstWeightPath(dw)], like -dw mxnet_converted-0000.param")
            emitter = MXNetEmitter((args.IRModelPath, args.IRWeightPath, args.dstWeightPath))
    elif args.dstFramework == 'onnx':
        from mmdnn.conversion.onnx.onnx_emitter import OnnxEmitter
        if args.IRWeightPath is None:
            raise NotImplementedError("ONNX emitter needs IR weight file")
        else:
            emitter = OnnxEmitter(args.IRModelPath, args.IRWeightPath)
    else:
        assert False

    emitter.run(args.dstModelPath, args.dstWeightPath, args.phase)

    return 0


def _get_parser():
    import argparse

    parser = argparse.ArgumentParser(description = 'Convert IR model file formats to other format.')

    parser.add_argument(
        '--phase',
        type=_text_type,
        choices=['train', 'test'],
        default='test',
        help='Convert phase (train/test) for destination toolkits.'
    )

    parser.add_argument(
        '--dstFramework', '-f',
        type=_text_type,
        choices=['caffe', 'caffe2', 'cntk', 'mxnet', 'keras', 'tensorflow', 'coreml', 'pytorch', 'onnx'],
        required=True,
        help='Format of model at srcModelPath (default is to auto-detect).')

    parser.add_argument(
        '--IRModelPath', '-n', '-in',
        type=_text_type,
        required=True,
        help='Path to the IR network structure file.')

    parser.add_argument(
        '--IRWeightPath', '-w', '-iw',
        type=_text_type,
        required=False,
        default=None,
        help = 'Path to the IR network structure file.')

    parser.add_argument(
        '--dstModelPath', '-d', '-o',
        type = _text_type,
        required = True,
        help = 'Path to save the destination model')

    # MXNet
    parser.add_argument(
        '--dstWeightPath', '-dw', '-ow',
        type=_text_type,
        default=None,
        help='[MXNet] Path to save the destination weight.')
    return parser


def _main():
    parser=_get_parser()
    args = parser.parse_args()
    ret = _convert(args)
    _sys.exit(int(ret)) # cast to int or else the exit code is always 1


if __name__ == '__main__':
    _main()


import caffe


def save_model(MainModel, network_filepath, weight_filepath, dump_filepath):
    dump_net = dump_filepath + '.prototxt'
    dump_weight = dump_filepath + '.caffemodel'
    dump_net = str(dump_net)
    dump_weight = str(dump_weight)
    MainModel.make_net(dump_net)
    MainModel.gen_weight(weight_filepath, dump_weight, dump_net)
    print('Caffe model files are saved as [{}] and [{}], generated by [{}.py] and [{}].'.format(
        dump_net, dump_weight, network_filepath, weight_filepath))
